from kivy_ios.toolchain import Recipe, shprint
from os.path import join
import sh


class FFMpegRecipe(Recipe):
    version = "n4.3.1"
    url = "https://github.com/FFmpeg/FFmpeg/archive/{version}.zip"
    include_per_arch = True
    include_dir = "dist/include"
    optional_depends = ["openssl"]
    libraries = [
        "libavcodec/libavcodec.a",
        "libavdevice/libavdevice.a",
        "libavfilter/libavfilter.a",
        "libavformat/libavformat.a",
        "libavresample/libavresample.a",
        "libavutil/libavutil.a",
        "libswresample/libswresample.a",
        "libswscale/libswscale.a",
    ]
    pbx_frameworks = ["VideoToolbox"]

    def build_arch(self, arch):
        options = [
            "--disable-everything",
            "--enable-parsers",
            "--enable-decoders",
            "--enable-demuxers",
            "--enable-filter=aresample,resample,crop,scale",
            "--enable-protocol=file,http,rtmp",
            "--enable-pic",
            "--enable-small",
            "--enable-hwaccels",
            "--enable-static",
            "--disable-shared",
            # libpostproc is GPL: https://ffmpeg.org/pipermail/ffmpeg-user/2012-February/005162.html
            "--enable-gpl",
            # disable some unused algo
            # note: "golomb" are the one used in our video test, so don't use --disable-golomb
            # note: and for aac decoding: "rdft", "mdct", and "fft" are needed
            "--disable-dxva2",
            "--disable-vdpau",
            "--disable-vaapi",
            "--disable-dct",
            # disable binaries / doc
            "--enable-cross-compile",
            "--disable-debug",
            "--disable-programs",
            "--disable-doc",
            "--enable-pic",
            "--enable-avresample"]

        if "openssl.build_all" in self.ctx.state:
            options += [
                "--enable-openssl",
                "--enable-nonfree",
                "--enable-protocol=https,tls_openssl"]

        build_env = arch.get_env()
        build_env["VERBOSE"] = "1"
        configure = sh.Command(join(self.build_dir, "configure"))
        shprint(configure,
                "--target-os=darwin",
                "--arch={}".format(arch.arch),
                "--cc={}".format(build_env["CC"]),
                "--prefix={}/dist".format(self.build_dir),
                "--extra-cflags={}".format(build_env["CFLAGS"]),
                "--extra-cxxflags={}".format(build_env["CFLAGS"]),
                "--extra-ldflags={}".format(build_env["LDFLAGS"]),
                "--disable-x86asm",
                *options,
                _env=build_env)
        """
        shprint(sh.sed,
                "-i.bak",
                "s/HAVE_CLOSESOCKET=yes//g",
                "config.mak")
        shprint(sh.sed,
                "-i.bak",
                "s/#define HAVE_CLOSESOCKET 1//g",
                "config.h")
        if exists("config.asm"):
            shprint(sh.sed,
                    "-i.bak",
                    "s/%define HAVE_CLOSESOCKET 1//g",
                    "config.asm")
        """
        shprint(sh.make, "clean", _env=build_env)
        shprint(sh.make, self.ctx.concurrent_make, _env=build_env)
        shprint(sh.make, "install")


recipe = FFMpegRecipe()



from kivy_ios.toolchain import CythonRecipe
from os.path import join
import logging
import shutil

logger = logging.getLogger(__name__)


class KivyRecipe(CythonRecipe):
    """
    The targeted commit includes these iOS specific fixes that were merged after kivy 2.0.0 release:
    - Camera: Added API to change avfoundation camera provider orientation (PR #7263)
    """
    version = "b5ec51ed2315a9a890f264b0df6d23c8e6341d42"
    url = "https://github.com/kivy/kivy/archive/{version}.zip"
    library = "libkivy.a"
    depends = ["sdl2", "sdl2_image", "sdl2_mixer", "sdl2_ttf", "ios",
               "pyobjus", "python", "host_setuptools3"]
    python_depends = ["certifi"]
    pbx_frameworks = ["OpenGLES", "Accelerate", "CoreMedia", "CoreVideo"]
    pre_build_ext = True

    def get_recipe_env(self, arch):
        env = super().get_recipe_env(arch)
        env["KIVY_SDL2_PATH"] = ":".join([
            join(self.ctx.dist_dir, "include", "common", "sdl2"),
            join(self.ctx.dist_dir, "include", "common", "sdl2_image"),
            join(self.ctx.dist_dir, "include", "common", "sdl2_ttf"),
            join(self.ctx.dist_dir, "include", "common", "sdl2_mixer")])
        return env

    def build_arch(self, arch):
        self._patch_setup()
        super().build_arch(arch)

    def _patch_setup(self):
        # patch setup to remove some functionnalities
        pyconfig = join(self.build_dir, "setup.py")

        def _remove_line(lines, pattern):
            for line in lines[:]:
                if pattern in line:
                    lines.remove(line)
        with open(pyconfig) as fd:
            lines = fd.readlines()
        _remove_line(lines, "flags['libraries'] = ['GLESv2']")
        with open(pyconfig, "w") as fd:
            fd.writelines(lines)

    def reduce_python_package(self):
        dest_dir = join(self.ctx.site_packages_dir, "kivy")
        shutil.rmtree(join(dest_dir, "tools"))
        shutil.rmtree(join(dest_dir, "tests"))


recipe = KivyRecipe()



from kivy_ios.toolchain import CythonRecipe
from os.path import join
import sh
import shutil


class NumpyRecipe(CythonRecipe):
    version = "1.16.4"
    url = "https://pypi.python.org/packages/source/n/numpy/numpy-{version}.zip"
    library = "libnumpy.a"
    libraries = ["libnpymath.a", "libnpysort.a"]
    include_dir = "numpy/core/include"
    depends = ["python"]
    pbx_frameworks = ["Accelerate"]
    cythonize = False

    def prebuild_arch(self, arch):
        if self.has_marker("patched"):
            return
        self.apply_patch("numpy-1.16.4.patch")
        self.set_marker("patched")

    def get_recipe_env(self, arch):
        env = super().get_recipe_env(arch)
        # CC must have the CFLAGS with arm arch, because numpy tries first to
        # compile and execute an empty C to see if the compiler works. This is
        # obviously not working when crosscompiling
        env["CC"] = "{} {}".format(env["CC"], env["CFLAGS"])
        # Numpy configuration. Don't try to compile anything related to it,
        # we're going to use the Accelerate framework
        env["NPYCONFIG"] = "env BLAS=None LAPACK=None ATLAS=None"
        return env

    def build_arch(self, arch):
        super().build_arch(arch)
        sh.cp(sh.glob(join(self.build_dir, "build", "temp.*", "libnpy*.a")),
              self.build_dir)

    def reduce_python_package(self):
        dest_dir = join(self.ctx.site_packages_dir, "numpy")
        shutil.rmtree(join(dest_dir, "core", "include"))
        shutil.rmtree(join(dest_dir, "core", "tests"))
        shutil.rmtree(join(dest_dir, "distutils"))
        shutil.rmtree(join(dest_dir, "doc"))
        shutil.rmtree(join(dest_dir, "f2py", "tests"))
        shutil.rmtree(join(dest_dir, "fft", "tests"))
        shutil.rmtree(join(dest_dir, "lib", "tests"))
        shutil.rmtree(join(dest_dir, "ma", "tests"))
        shutil.rmtree(join(dest_dir, "matrixlib", "tests"))
        shutil.rmtree(join(dest_dir, "polynomial", "tests"))
        shutil.rmtree(join(dest_dir, "random", "tests"))
        shutil.rmtree(join(dest_dir, "tests"))


recipe = NumpyRecipe()



from kivy_ios.toolchain import Recipe, shprint
from os.path import join
import os
import sh
import shutil
import logging
from kivy_ios.context_managers import cd


logger = logging.getLogger(__name__)


class Hostpython3Recipe(Recipe):
    version = "3.8.2"
    url = "https://www.python.org/ftp/python/{version}/Python-{version}.tgz"
    depends = ["hostlibffi", "hostopenssl"]
    optional_depends = []
    archs = ["x86_64"]
    build_subdir = 'native-build'

    def init_with_ctx(self, ctx):
        super().init_with_ctx(ctx)
        self.set_hostpython(self, "3.8")
        self.ctx.so_suffix = ".cpython-38m-darwin.so"
        self.ctx.hostpython = join(self.ctx.dist_dir, "hostpython3", "bin", "python")
        self.ctx.hostpgen = join(self.ctx.dist_dir, "hostpython3", "bin", "pgen")
        logger.info("Global: hostpython located at {}".format(self.ctx.hostpython))
        logger.info("Global: hostpgen located at {}".format(self.ctx.hostpgen))

    def get_build_subdir(self, arch):
        return join(self.get_build_dir(arch), self.build_subdir)

    def prebuild_arch(self, arch):
        if self.has_marker("patched"):
            return
        self.apply_patch("pyconfig_detection.patch")
        self.copy_file("ModulesSetup", "Modules/Setup.local")
        self.set_marker("patched")

    def postbuild_arch(self, arch):
        return

    def get_build_env(self):
        sdk_path = sh.xcrun("--sdk", "macosx", "--show-sdk-path").strip()
        build_env = self.ctx.env.copy()
        self.build_env_x86_84 = build_env
        ccache = (build_env["CCACHE"] + ' ') if 'CCACHE' in build_env else ''
        build_env["CC"] = ccache + "clang -Qunused-arguments -fcolor-diagnostics"
        build_env["LDFLAGS"] = " ".join([
                "-lsqlite3",
                "-lffi",
                "-L{}".format(join(self.ctx.dist_dir, "hostlibffi", "usr", "local", "lib"))
                ])
        build_env["CFLAGS"] = " ".join([
                "--sysroot={}".format(sdk_path),
                "-arch x86_64",
                "-mmacosx-version-min=10.12",
                "-I{}".format(join(self.ctx.dist_dir, "hostlibffi", "usr", "local", "include"))
                ])
        return build_env

    def build_x86_64(self):
        build_env = self.get_build_env()
        configure = sh.Command(join(self.build_dir, "configure"))
        arch = self.filtered_archs[0]
        build_subdir = self.get_build_subdir(arch.arch)
        os.makedirs(build_subdir, exist_ok=True)
        with cd(build_subdir):
            shprint(configure,
                    "ac_cv_func_preadv=no",
                    "ac_cv_func_pwritev=no",
                    "--prefix={}".format(join(self.ctx.dist_dir, "hostpython3")),
                    "--with-openssl={}".format(join(self.ctx.dist_dir, 'hostopenssl')),
                    _env=build_env)
        shprint(sh.make, "-C", build_subdir, self.ctx.concurrent_make,
                _env=build_env)

    def install(self):
        arch = list(self.filtered_archs)[0]
        build_env = self.get_build_env()
        build_subdir = self.get_build_subdir(arch.arch)
        build_env["PATH"] = os.environ["PATH"]
        shprint(sh.make, self.ctx.concurrent_make,
                "-C", build_subdir,
                "install",
                _env=build_env)
        shutil.copy(
            join(self.ctx.dist_dir, "hostpython3", "bin", "python3"),
            join(self.ctx.dist_dir, "hostpython3", "bin", "python"))
        """
        I don't like this kind of "patches".
        sysconfig was overriding our cflags and extensions were failing to build.
        This hack resets the cflags provided by sysconfig.
        """
        with open(join(self.ctx.dist_dir, "hostpython3", "lib", "python3.8", "distutils", "sysconfig.py"), 'r') as sysconfigfile:
            lines = sysconfigfile.readlines()
        lines[192] = '        cflags = ""\n'
        with open(join(self.ctx.dist_dir, "hostpython3", "lib", "python3.8", "distutils", "sysconfig.py"), 'w') as sysconfigfile:
            sysconfigfile.writelines(lines)


recipe = Hostpython3Recipe()



# -*- coding: utf-8 -*-
from kivy_ios.toolchain import Recipe
import logging
from os.path import join

logger = logging.getLogger(__name__)


class PythonAliasRecipe(Recipe):
    """
    Note this recipe was created to handle both python2 and python3.
    As python2 support was dropped, this could probably be simplified.
    """
    is_alias = True

    def init_after_import(self, ctx):
        python = ctx.state.get("python")
        if not python:
            python = "python3"
        self.depends = [python]
        self.recipe_dir = join(ctx.root_dir, "recipes", python)


recipe = PythonAliasRecipe()



from kivy_ios.toolchain import Recipe, shprint
from os.path import join
import sh


class CurlRecipe(Recipe):
    version = "7.65.3"
    url = "https://curl.haxx.se/download/curl-{version}.tar.gz"
    library = "lib/.libs/libcurl.a"
    include_dir = "include"
    depends = ["openssl"]

    def build_arch(self, arch):
        build_env = arch.get_env()
        configure = sh.Command(join(self.build_dir, "configure"))
        shprint(configure,
                "CC={}".format(build_env["CC"]),
                "LD={}".format(build_env["LD"]),
                "CFLAGS={}".format(build_env["CFLAGS"]),
                "LDFLAGS={}".format(build_env["LDFLAGS"]),
                "--prefix=/",
                "--host={}".format(arch.triple),
                "--disable-shared",
                "--without-libidn2")
        shprint(sh.make, "clean")
        shprint(sh.make, self.ctx.concurrent_make)


recipe = CurlRecipe()



from kivy_ios.toolchain import Recipe, shprint
import sh


class LibSDL2MixerRecipe(Recipe):
    version = "2.0.4"
    url = "http://www.libsdl.org/projects/SDL_mixer/release/SDL2_mixer-{version}.tar.gz"
    library = "Xcode-iOS/build/Release-{arch.sdk}/libSDL2_mixer.a"
    include_dir = "SDL_mixer.h"
    depends = ["sdl2"]
    pbx_frameworks = ["ImageIO"]
    pbx_libraries = ["libc++"]

    def build_arch(self, arch):
        # endian.h is in /usr/include/machine/ (Since MacOs Mojave?)
        # header is needed by libvorbis, so We're adding that folder
        # to HEADER_SEARCH_PATHS
        shprint(sh.xcodebuild, self.ctx.concurrent_xcodebuild,
                "ONLY_ACTIVE_ARCH=NO",
                "ARCHS={}".format(arch.arch),
                "HEADER_SEARCH_PATHS=$HEADER_SEARCH_PATHS /usr/include/machine {} ".format(" ".join(arch.include_dirs)),
                "-sdk", arch.sdk,
                "-project", "Xcode-iOS/SDL_mixer.xcodeproj",
                "-target", "libSDL_mixer-iOS",
                "-configuration", "Release")


recipe = LibSDL2MixerRecipe()



from kivy_ios.toolchain import CythonRecipe


class PyobjusRecipe(CythonRecipe):
    version = "master"
    url = "https://github.com/kivy/pyobjus/archive/{version}.zip"
    library = "libpyobjus.a"
    depends = ["python"]
    pre_build_ext = True


recipe = PyobjusRecipe()



import sh

from os.path import join

from kivy_ios.toolchain import CythonRecipe, shprint
from kivy_ios.context_managers import cd


class NetifacesRecipe(CythonRecipe):
    version = "0.10.9"
    url = "https://pypi.io/packages/source/n/netifaces/netifaces-{version}.tar.gz"
    depends = ["python3", "host_setuptools3"]
    python_depends = ["setuptools"]
    library = "libnetifaces.a"
    cythonize = False

    def dest_dir(self):
        return join(self.ctx.dist_dir, "root", "python3")

    def get_netifaces_env(self, arch):
        build_env = arch.get_env()
        build_env["PYTHONPATH"] = join(
            self.dest_dir(), "lib", "python3.8", "site-packages"
        )
        return build_env

    def install(self):
        arch = list(self.filtered_archs)[0]
        build_dir = self.get_build_dir(arch.arch)
        build_env = self.get_netifaces_env(arch)
        hostpython = sh.Command(self.ctx.hostpython)
        with cd(build_dir):
            shprint(
                hostpython,
                "setup.py",
                "install",
                "--prefix",
                self.dest_dir(),
                _env=build_env,
            )


recipe = NetifacesRecipe()



from kivy_ios.toolchain import PythonRecipe


class PykkaRecipe(PythonRecipe):
    version = '1.2.1'
    url = 'https://github.com/jodal/pykka/archive/v{version}.zip'

    depends = ['python', 'host_setuptools']

    site_packages_name = 'pykka'


recipe = PykkaRecipe()



# pure-python package, this can be removed when we'll support any python package
from kivy_ios.toolchain import PythonRecipe, shprint
from os.path import join
import sh
import os


class FlaskRecipe(PythonRecipe):
    version = "1.1.2"
    url = "https://github.com/mitsuhiko/flask/archive/{version}.zip"
    depends = ["python", "jinja2", "werkzeug", "itsdangerous", "click"]

    def install(self):
        arch = list(self.filtered_archs)[0]
        build_dir = self.get_build_dir(arch.arch)
        os.chdir(build_dir)
        hostpython = sh.Command(self.ctx.hostpython)
        build_env = arch.get_env()
        dest_dir = join(self.ctx.dist_dir, "root", "python")
        build_env['PYTHONPATH'] = join(dest_dir, 'lib', 'python3.8', 'site-packages')
        shprint(hostpython, "setup.py", "install", "--prefix", dest_dir, _env=build_env)


recipe = FlaskRecipe()



from kivy_ios.toolchain import Recipe, shprint
from os.path import join
import sh
import os
import shutil


class HostSetuptools(Recipe):
    depends = ["openssl", "hostpython"]
    archs = ["x86_64"]
    url = "setuptools"

    def prebuild_arch(self, arch):
        hostpython = sh.Command(self.ctx.hostpython)
        sh.curl("-O", "https://bootstrap.pypa.io/ez_setup.py")
        shprint(hostpython, "./ez_setup.py")
        # Extract setuptools egg and remove .pth files. Otherwise subsequent
        # python package installations using setuptools will raise exceptions.
        # Setuptools version 28.3.0
        site_packages_path = join(
            self.ctx.dist_dir, 'hostpython',
            'lib', 'python3.7', 'site-packages')
        os.chdir(site_packages_path)
        with open('setuptools.pth', 'r') as f:
            setuptools_egg_path = f.read().strip('./').strip('\n')
            unzip = sh.Command('unzip')
            shprint(unzip, setuptools_egg_path)
        os.remove(setuptools_egg_path)
        os.remove('setuptools.pth')
        os.remove('easy-install.pth')
        shutil.rmtree('EGG-INFO')


recipe = HostSetuptools()



from kivy_ios.toolchain import CythonRecipe


class AudiostreamRecipe(CythonRecipe):
    version = "master"
    url = "https://github.com/kivy/audiostream/archive/{version}.zip"
    library = "libaudiostream.a"
    depends = ["python", "sdl2", "sdl2_mixer"]
    pre_build_ext = True


recipe = AudiostreamRecipe()



from kivy_ios.toolchain import Recipe, shprint, cache_execution
from kivy_ios.context_managers import cd, python_path
import sh


class HostSetuptools3(Recipe):
    depends = ["openssl", "hostpython3", "python3"]
    archs = ["x86_64"]
    version = '40.9.0'
    url = 'https://pypi.python.org/packages/source/s/setuptools/setuptools-{version}.zip'

    @cache_execution
    def install(self):
        arch = self.filtered_archs[0]
        build_dir = self.get_build_dir(arch.arch)
        hostpython = sh.Command(self.ctx.hostpython)

        with python_path(self.ctx.site_packages_dir):
            with cd(build_dir):
                shprint(hostpython, "setup.py", "install",
                        f"--prefix={self.ctx.python_prefix}")


recipe = HostSetuptools3()



from kivy_ios.toolchain import Recipe, shprint
import sh
from os.path import exists
import logging

logger = logging.getLogger(__name__)


class LibffiRecipe(Recipe):
    version = "3.2.1"
    url = "ftp://sourceware.org/pub/libffi/libffi-{version}.tar.gz"
    library = "build/Release-{arch.sdk}/libffi.a"
    include_per_arch = True
    include_dir = "build_{arch.sdk}-{arch.arch}/include"
    archs = ["x86_64"]

    def build_all(self):
        filtered_archs = self.filtered_archs
        logger.info("Build {} for {} (filtered)".format(
            self.name,
            ", ".join([x.arch for x in filtered_archs])))
        for arch in self.filtered_archs:
            self.build(arch)

        # since we don't run cache_execution, call this here for `status`
        self.update_state("{}.build_all".format(self.name), True)

    def prebuild_arch(self, arch):
        if self.has_marker("patched"):
            return
        # necessary as it doesn't compile with XCode 6.0. If we use 5.1.1, the
        # compiler for i386 is not working.
        # shprint(sh.sed,
        #        "-i.bak",
        #        "s/-miphoneos-version-min=5.1.1/-miphoneos-version-min=6.0/g",
        #        "generate-darwin-source-and-headers.py")
        self.apply_patch("fix-win32-unreferenced-symbol.patch")
        self.apply_patch("public_include.patch")
        self.apply_patch("staticlib.patch")
        self.apply_patch("staticlib2.patch")
        self.apply_patch("libffi-xcode10.patch")
        self.apply_patch("generate-darwin-source-and-headers-python3-items.patch")
        self.set_marker("patched")

    def build_arch(self, arch):
        if exists("generate-darwin-source-and-headers.py"):
            shprint(
                sh.mv,
                "generate-darwin-source-and-headers.py",
                "_generate-darwin-source-and-headers.py")
            shprint(sh.touch, "generate-darwin-source-and-headers.py")
        python3 = sh.Command("python3")
        shprint(python3, "_generate-darwin-source-and-headers.py", "--only-osx")
        shprint(sh.xcodebuild,
                self.ctx.concurrent_xcodebuild,
                "ONLY_ACTIVE_ARCH=NO",
                "ARCHS={}".format(arch.arch),
                "DSTROOT={}/hostlibffi".format(self.ctx.dist_dir),
                "-sdk", "macosx",
                "clean", "build", "installhdrs", "install",
                "-project", "libffi.xcodeproj",
                "-scheme", "libffi-Mac",
                "-configuration", "Release")

    def postbuild_arch(self, arch):
        pass


recipe = LibffiRecipe()



from kivy_ios.toolchain import Recipe, shprint
from kivy_ios.context_managers import cd
from os.path import join
import sh
import shutil
import os
import logging

logger = logging.getLogger(__name__)


class Python3Recipe(Recipe):
    version = "3.8.2"
    url = "https://www.python.org/ftp/python/{version}/Python-{version}.tgz"
    depends = ["hostpython3", "libffi", "openssl"]
    library = "libpython3.8.a"
    pbx_libraries = ["libz", "libbz2", "libsqlite3"]

    def init_with_ctx(self, ctx):
        super().init_with_ctx(ctx)
        self.set_python(self, "3.8")
        ctx.python_ver_dir = "python3.8"
        ctx.python_prefix = join(ctx.dist_dir, "root", "python3")
        ctx.site_packages_dir = join(
            ctx.python_prefix, "lib", ctx.python_ver_dir, "site-packages")

    def prebuild_arch(self, arch):
        # common to all archs
        if self.has_marker("patched"):
            return
        self.apply_patch("config.sub.patch")
        self.apply_patch("configure.patch")
        self.apply_patch("posixmodule.patch")
        self.apply_patch("dynload_shlib.patch")
        self.apply_patch("disable_explicit_blake2.patch")
        self.apply_patch("pyconfig_detection.patch")
        self.copy_file("ModulesSetup", "Modules/Setup.local")
        self.append_file("ModulesSetup.mobile", "Modules/Setup.local")
        self.set_marker("patched")

    def postbuild_arch(self, arch):
        # include _sqlite module to .a
        py_arch = arch.arch
        if py_arch == "armv7":
            py_arch = "arm"
        elif py_arch == "arm64":
            py_arch = "aarch64"
        tmp_folder = "temp.ios-{}-3.8{}".format(py_arch, self.build_dir)
        build_env = self.get_build_env(arch)
        for o_file in [
            "cache.o",
            "cursor.o",
            "module.o",
            "row.o",
            "util.o",
            "connection.o",
            "microprotocols.o",
            "prepare_protocol.o",
            "statement.o",
        ]:
            shprint(sh.Command(build_env['AR']),
                    "-r",
                    "{}/{}".format(self.build_dir, self.library),
                    "{}/build/{}/Modules/_sqlite/{}".format(self.build_dir, tmp_folder, o_file))
        print("Added _sqlite to archive")

    def get_build_env(self, arch):
        build_env = arch.get_env()
        build_env["PATH"] = "{}:{}".format(
            join(self.ctx.dist_dir, "hostpython3", "bin"),
            os.environ["PATH"])
        build_env["CFLAGS"] += " --sysroot={}".format(arch.sysroot)
        return build_env

    def build_arch(self, arch):
        build_env = self.get_build_env(arch)
        configure = sh.Command(join(self.build_dir, "configure"))
        py_arch = arch.arch
        if py_arch == "armv7":
            py_arch = "arm"
        elif py_arch == "arm64":
            py_arch = "aarch64"
        prefix = join(self.ctx.dist_dir, "root", "python3")
        shprint(configure,
                "CC={}".format(build_env["CC"]),
                "LD={}".format(build_env["LD"]),
                "CFLAGS={}".format(build_env["CFLAGS"]),
                "LDFLAGS={} -undefined dynamic_lookup".format(build_env["LDFLAGS"]),
                "ac_cv_file__dev_ptmx=yes",
                "ac_cv_file__dev_ptc=no",
                "ac_cv_little_endian_double=yes",
                "ac_cv_func_memrchr=no",
                "ac_cv_func_getentropy=no",
                "ac_cv_func_getresuid=no",
                "ac_cv_func_getresgid=no",
                "ac_cv_func_setresgid=no",
                "ac_cv_func_setresuid=no",
                "ac_cv_func_plock=no",
                "ac_cv_func_dup3=no",
                "ac_cv_func_pipe2=no",
                "ac_cv_func_preadv=no",
                "ac_cv_func_pwritev=no",
                "ac_cv_func_preadv2=no",
                "ac_cv_func_pwritev2=no",
                "ac_cv_func_mkfifoat=no",
                "ac_cv_func_mknodat=no",
                "ac_cv_func_posix_fadvise=no",
                "ac_cv_func_posix_fallocate=no",
                "ac_cv_func_sigwaitinfo=no",
                "ac_cv_func_sigtimedwait=no",
                "ac_cv_func_clock_settime=no",
                "ac_cv_func_pthread_getcpuclockid=no",
                "ac_cv_func_sched_setscheduler=no",
                "ac_cv_func_sched_setparam=no",
                "ac_cv_func_clock_gettime=no",
                "--host={}-apple-ios".format(py_arch),
                "--build=x86_64-apple-darwin",
                "--prefix={}".format(prefix),
                "--without-ensurepip",
                "--with-system-ffi",
                "--enable-ipv6",
                "PYTHON_FOR_BUILD=_PYTHON_PROJECT_BASE=$(abs_builddir) \
                    _PYTHON_HOST_PLATFORM=$(_PYTHON_HOST_PLATFORM) \
                    PYTHONPATH=$(shell test -f pybuilddir.txt && echo $(abs_builddir)/`cat pybuilddir.txt`:)$(srcdir)/Lib\
                    _PYTHON_SYSCONFIGDATA_NAME=_sysconfigdata_$(ABIFLAGS)_$(MACHDEP)_$(MULTIARCH)\
                    {}".format(sh.Command(self.ctx.hostpython)),
                _env=build_env)
        self.apply_patch("ctypes_duplicate.patch")
        shprint(sh.make, self.ctx.concurrent_make)

    def install(self):
        arch = list(self.filtered_archs)[0]
        build_env = self.get_build_env(arch)
        build_dir = self.get_build_dir(arch.arch)
        shprint(sh.make, self.ctx.concurrent_make,
                "-C", build_dir,
                "install",
                "prefix={}".format(join(self.ctx.dist_dir, "root", "python3")),
                _env=build_env)
        self.reduce_python()
        self.install_mock_modules()

    def install_mock_modules(self):
        logger.info("Install mock modules")
        sqlite3_src = join(self.recipe_dir, 'mock_modules', '_sqlite3')
        site_packages_folder = join(
                self.ctx.dist_dir, "root", "python3", "lib", "python3.8", "site-packages", "_sqlite3")
        shutil.rmtree(site_packages_folder, ignore_errors=True)  # Needed in case of rebuild
        shutil.copytree(sqlite3_src, site_packages_folder)

    def reduce_python(self):
        logger.info("Reduce python")
        logger.info("Remove files unlikely to be used")
        with cd(join(self.ctx.dist_dir, "root", "python3")):
            sh.rm("-rf", "bin", "share")
        # platform binaries and configuration
        with cd(join(
                self.ctx.dist_dir, "root", "python3", "lib",
                "python3.8", "config-3.8-darwin")):
            sh.rm(
                "libpython3.8.a",
                "python.o",
                "config.c.in",
                "makesetup",
                "install-sh",
            )

        # cleanup pkgconfig and compiled lib
        with cd(join(self.ctx.dist_dir, "root", "python3", "lib")):
            sh.rm("-rf", "pkgconfig", "libpython3.8.a")

        # cleanup python libraries
        with cd(join(
                self.ctx.dist_dir, "root", "python3", "lib", "python3.8")):
            sh.rm("-rf", "wsgiref", "curses", "idlelib", "lib2to3",
                  "ensurepip", "turtledemo", "lib-dynload", "venv",
                  "pydoc_data")
            sh.find(".", "-path", "*/test*/*", "-delete")
            sh.find(".", "-name", "*.exe", "-type", "f", "-delete")
            sh.find(".", "-name", "test*", "-type", "d", "-delete")
            sh.find(".", "-iname", "*.pyc", "-delete")
            sh.find(".", "-path", "*/__pycache__/*", "-delete")
            sh.find(".", "-name", "__pycache__", "-type", "d", "-delete")

            # now precompile to Python bytecode
            hostpython = sh.Command(self.ctx.hostpython)
            shprint(hostpython, "-m", "compileall", "-f", "-b")
            # sh.find(".", "-iname", "*.py", "-delete")

            # some pycache are recreated after compileall
            sh.find(".", "-path", "*/__pycache__/*", "-delete")
            sh.find(".", "-name", "__pycache__", "-type", "d", "-delete")

            # create the lib zip
            logger.info("Create a python3.8.zip")
            sh.mv("config-3.8-darwin", "..")
            sh.mv("site-packages", "..")
            sh.zip("-r", "../python38.zip", sh.glob("*"))
            sh.rm("-rf", sh.glob("*"))
            sh.mv("../config-3.8-darwin", ".")
            sh.mv("../site-packages", ".")


recipe = Python3Recipe()



# pure-python package, this can be removed when we'll support any python package
from kivy_ios.toolchain import PythonRecipe, shprint
from os.path import join
import sh
import os


class MarkupSafeRecipe(PythonRecipe):
    version = "1.1.1"
    url = "https://github.com/mitsuhiko/markupsafe/archive/{version}.zip"
    depends = ["python"]

    def install(self):
        arch = list(self.filtered_archs)[0]
        build_dir = self.get_build_dir(arch.arch)
        os.chdir(build_dir)
        hostpython = sh.Command(self.ctx.hostpython)
        build_env = arch.get_env()
        dest_dir = join(self.ctx.dist_dir, "root", "python3")
        build_env['PYTHONPATH'] = join(dest_dir, 'lib', 'python3.8', 'site-packages')
        cmd = sh.Command("sed")
        shprint(cmd, "-i", "", "s/,.*Feature//g", "./setup.py", _env=build_env)
        shprint(cmd, "-i", "", "/^speedups = Feature/,/^)$/s/.*//g", "./setup.py", _env=build_env)
        shprint(cmd, "-i", "", "s/features\['speedups'\].*=.*speedups/pass/g", "./setup.py", _env=build_env)  # noqa: W605
        shprint(hostpython, "setup.py", "install", "--prefix", dest_dir, _env=build_env)


recipe = MarkupSafeRecipe()



from distutils.core import setup, Extension


setup(name='ios',
      version='1.1',
      ext_modules=[
          Extension('ios',
                    ['ios.c', 'ios_utils.m', 'ios_mail.m', 'ios_browser.m',
                     'ios_filechooser.m'],
                    libraries=[],
                    library_dirs=[],
                    )
                  ]
      )



from kivy_ios.toolchain import PythonRecipe


class PlyerRecipe(PythonRecipe):
    version = "master"
    url = "https://github.com/kivy/plyer/archive/{version}.zip"
    depends = ["python", "pyobjus"]
    archs = ["x86_64"]


recipe = PlyerRecipe()



#!/usr/bin/env python

import os
import sys
import subprocess

# resolve cython executable
cython = None


def resolve_cython():
    global cython
    for executable in ('cython', 'cython-2.7'):
        for path in os.environ['PATH'].split(':'):
            if not os.path.exists(path):
                continue
            if executable in os.listdir(path):
                cython = os.path.join(path, executable)
                return


def do(fn):
    print('cythonize:', fn)
    assert(fn.endswith('.pyx'))
    parts = fn.split('/')
    if parts[0] == '.':
        parts.pop(0)
    modname = parts[-1][:-4]
    package = '_'.join(parts[:-1])

    # cythonize
    subprocess.Popen([cython, fn], env=os.environ).communicate()

    if not package:
        print('no need to rewrite', fn)
    else:
        # get the .c, and change the initXXX
        fn_c = fn[:-3] + 'c'
        with open(fn_c) as fd:
            data = fd.read()
        modname = modname.split('.')[-1]
        pac_mod = '{}_{}'.format(package, modname)
        fmts = ('init{}(void)', 'PyInit_{}(void)', 'Pyx_NAMESTR("{}")', '"{}",')
        for i, fmt in enumerate(fmts):
            pat = fmt.format(modname)
            sub = fmt.format(pac_mod)
            print('{}: {} -> {}'.format(i + 1, pat, sub))
            data = data.replace(pat, sub)
        print('rewrite', fn_c)
        with open(fn_c, 'w') as fd:
            fd.write(data)


if __name__ == '__main__':
    print('-- cythonize', sys.argv)
    resolve_cython()
    for fn in sys.argv[1:]:
        do(fn)



# pure-python package, this can be removed when we'll support any python package
import os
import sh
from kivy_ios.toolchain import PythonRecipe, shprint


class PyYamlRecipe(PythonRecipe):
    version = "3.11"
    url = "https://pypi.python.org/packages/source/P/PyYAML/PyYAML-{version}.tar.gz"
    depends = ["python"]

    def install(self):
        arch = list(self.filtered_archs)[0]
        build_dir = self.get_build_dir(arch.arch)
        os.chdir(build_dir)
        hostpython = sh.Command(self.ctx.hostpython)
        build_env = arch.get_env()
        dest_dir = os.path.join(self.ctx.dist_dir, "root", "python")
        build_env['PYTHONPATH'] = os.path.join(dest_dir, 'lib', 'python3.7', 'site-packages')
        shprint(hostpython, "setup.py", "install", "--prefix", dest_dir, _env=build_env)


recipe = PyYamlRecipe()



#!/usr/bin/env python
"""
Tool for compiling iOS toolchain
================================

This tool intend to replace all the previous tools/ in shell script.
"""

import argparse
import sys
from sys import stdout
from os.path import join, dirname, realpath, exists, isdir, basename
from os import listdir, unlink, makedirs, environ, chdir, getcwd, walk
import sh
import zipfile
import tarfile
import importlib
import io
import json
import shutil
import fnmatch
import tempfile
import time
from contextlib import suppress
from datetime import datetime
from pprint import pformat
import logging
from urllib.request import FancyURLopener, urlcleanup
from pbxproj import XcodeProject
from pbxproj.pbxextensions.ProjectFiles import FileOptions

curdir = dirname(__file__)

initial_working_directory = getcwd()

# For more detailed logging, use something like
# format='%(asctime)s,%(msecs)d %(levelname)-8s [%(filename)s:%(funcName)s():%(lineno)d] %(message)s'
logging.basicConfig(format='[%(levelname)-8s] %(message)s',
                    datefmt='%Y-%m-%d:%H:%M:%S',
                    level=logging.DEBUG)

# Quiet the loggers we don't care about
sh_logging = logging.getLogger('sh')
sh_logging.setLevel(logging.WARNING)

logger = logging.getLogger(__name__)


def shprint(command, *args, **kwargs):
    kwargs["_iter"] = True
    kwargs["_out_bufsize"] = 1
    kwargs["_err_to_out"] = True
    logger.info("Running Shell: {} {} {}".format(str(command), args, kwargs))
    cmd = command(*args, **kwargs)
    for line in cmd:
        # strip only last CR:
        line_str = "\n".join(line.encode("ascii", "replace").decode().splitlines())
        logger.debug(line_str)


def cache_execution(f):
    def _cache_execution(self, *args, **kwargs):
        state = self.ctx.state
        key = "{}.{}".format(self.name, f.__name__)
        force = kwargs.pop("force", False)
        if args:
            for arg in args:
                key += ".{}".format(arg)
        if key in state and not force:
            logger.debug("Cached result: {} {}. Ignoring".format(f.__name__.capitalize(), self.name))
            return
        logger.info("{} {}".format(f.__name__.capitalize(), self.name))
        f(self, *args, **kwargs)
        self.update_state(key, True)
    return _cache_execution


def remove_junk(d):
    """ Remove unused build artifacts. """
    exts = (".so.lib", ".so.o", ".sh")
    for root, dirnames, filenames in walk(d):
        for fn in filenames:
            if fn.endswith(exts):
                print('Found junk {}/{}, removing'.format(root, fn))
                unlink(join(root, fn))


class ChromeDownloader(FancyURLopener):
    version = (
        'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 '
        '(KHTML, like Gecko) Chrome/28.0.1500.71 Safari/537.36')


urlretrieve = ChromeDownloader().retrieve


class JsonStore:
    """Replacement of shelve using json, needed for support python 2 and 3.
    """

    def __init__(self, filename):
        self.filename = filename
        self.data = {}
        if exists(filename):
            try:
                with io.open(filename, encoding='utf-8') as fd:
                    self.data = json.load(fd)
            except ValueError:
                logger.warning("Unable to read the state.db, content will be replaced.")

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value
        self.sync()

    def __delitem__(self, key):
        del self.data[key]
        self.sync()

    def __contains__(self, item):
        return item in self.data

    def get(self, item, default=None):
        return self.data.get(item, default)

    def keys(self):
        return self.data.keys()

    def remove_all(self, prefix):
        for key in tuple(self.data.keys()):
            if not key.startswith(prefix):
                continue
            del self.data[key]
        self.sync()

    def sync(self):
        with open(self.filename, 'w') as fd:
            json.dump(self.data, fd, ensure_ascii=False)


class Arch:
    def __init__(self, ctx):
        self.ctx = ctx
        self._ccsh = None

    def __str__(self):
        return self.arch

    @property
    def include_dirs(self):
        return [
            "{}/{}".format(
                self.ctx.include_dir,
                d.format(arch=self))
            for d in self.ctx.include_dirs]

    def get_env(self):
        include_dirs = [
            "-I{}/{}".format(
                self.ctx.include_dir,
                d.format(arch=self))
            for d in self.ctx.include_dirs]
        include_dirs += ["-I{}".format(
            join(self.ctx.dist_dir, "include", self.arch))]

        env = {}
        cc = sh.xcrun("-find", "-sdk", self.sdk, "clang").strip()
        cxx = sh.xcrun("-find", "-sdk", self.sdk, "clang++").strip()

        # we put the flags in CC / CXX as sometimes the ./configure test
        # with the preprocessor (aka CC -E) without CFLAGS, which fails for
        # cross compiled projects
        flags = " ".join([
            "--sysroot", self.sysroot,
            "-arch", self.arch,
            "-pipe", "-no-cpp-precomp",
        ])
        cc += " " + flags
        cxx += " " + flags

        use_ccache = environ.get("USE_CCACHE", "1")
        ccache = None
        if use_ccache == "1":
            ccache = sh.which('ccache')
        if ccache:
            ccache = ccache.strip()
            env["USE_CCACHE"] = "1"
            env["CCACHE"] = ccache
            env.update({k: v for k, v in environ.items() if k.startswith('CCACHE_')})
            env.setdefault('CCACHE_MAXSIZE', '10G')
            env.setdefault('CCACHE_HARDLINK', 'true')
            env.setdefault(
                'CCACHE_SLOPPINESS',
                ('file_macro,time_macros,'
                 'include_file_mtime,include_file_ctime,file_stat_matches'))

        if not self._ccsh:
            self._ccsh = tempfile.NamedTemporaryFile()
            self._cxxsh = tempfile.NamedTemporaryFile()
            sh.chmod("+x", self._ccsh.name)
            sh.chmod("+x", self._cxxsh.name)
            self._ccsh.write(b'#!/bin/sh\n')
            self._cxxsh.write(b'#!/bin/sh\n')
            if ccache:
                logger.info("CC and CXX will use ccache")
                self._ccsh.write(
                    (ccache + ' ' + cc + ' "$@"\n').encode("utf8"))
                self._cxxsh.write(
                    (ccache + ' ' + cxx + ' "$@"\n').encode("utf8"))
            else:
                logger.info("CC and CXX will not use ccache")
                self._ccsh.write(
                    (cc + ' "$@"\n').encode("utf8"))
                self._cxxsh.write(
                    (cxx + ' "$@"\n').encode("utf8"))
            self._ccsh.flush()
            self._cxxsh.flush()

        env["CC"] = self._ccsh.name
        env["CXX"] = self._cxxsh.name
        env["AR"] = sh.xcrun("-find", "-sdk", self.sdk, "ar").strip()
        env["LD"] = sh.xcrun("-find", "-sdk", self.sdk, "ld").strip()
        env["OTHER_CFLAGS"] = " ".join(include_dirs)
        env["OTHER_LDFLAGS"] = " ".join([
            "-L{}/{}".format(self.ctx.dist_dir, "lib"),
        ])
        env["CFLAGS"] = " ".join([
            "-O3",
            self.version_min
        ] + include_dirs)
        env["LDFLAGS"] = " ".join([
            "-arch", self.arch,
            # "--sysroot", self.sysroot,
            "-L{}/{}".format(self.ctx.dist_dir, "lib"),
            "-L{}/usr/lib".format(self.sysroot),
            self.version_min
        ])
        return env


class Arch64Simulator(Arch):
    sdk = "iphonesimulator"
    arch = "x86_64"
    triple = "x86_64-apple-darwin13"
    version_min = "-miphoneos-version-min=8.0"
    sysroot = sh.xcrun("--sdk", "iphonesimulator", "--show-sdk-path").strip()


class Arch64IOS(Arch):
    sdk = "iphoneos"
    arch = "arm64"
    triple = "aarch64-apple-darwin13"
    version_min = "-miphoneos-version-min=8.0"
    sysroot = sh.xcrun("--sdk", "iphoneos", "--show-sdk-path").strip()


class Graph:
    # Taken from python-for-android/depsort
    def __init__(self):
        # `graph`: dict that maps each package to a set of its dependencies.
        self.graph = {}

    def add(self, dependent, dependency):
        """Add a dependency relationship to the graph"""
        self.graph.setdefault(dependent, set())
        self.graph.setdefault(dependency, set())
        if dependent != dependency:
            self.graph[dependent].add(dependency)

    def add_optional(self, dependent, dependency):
        """Add an optional (ordering only) dependency relationship to the graph

        Only call this after all mandatory requirements are added
        """
        if dependent in self.graph and dependency in self.graph:
            self.add(dependent, dependency)

    def find_order(self):
        """Do a topological sort on a dependency graph

        :Parameters:
            :Returns:
                iterator, sorted items form first to last
        """
        graph = dict((k, set(v)) for k, v in self.graph.items())
        while graph:
            # Find all items without a parent
            leftmost = [name for name, dep in graph.items() if not dep]
            if not leftmost:
                raise ValueError('Dependency cycle detected! %s' % graph)
            # If there is more than one, sort them for predictable order
            leftmost.sort()
            for result in leftmost:
                # Yield and remove them from the graph
                yield result
                graph.pop(result)
                for bset in graph.values():
                    bset.discard(result)


class Context:
    env = environ.copy()
    root_dir = None
    cache_dir = None
    build_dir = None
    dist_dir = None
    install_dir = None
    ccache = None
    cython = None
    sdkver = None
    sdksimver = None
    so_suffix = None  # set by one of the hostpython

    def __init__(self):
        self.include_dirs = []

        ok = True

        sdks = sh.xcodebuild("-showsdks").splitlines()

        # get the latest iphoneos
        iphoneos = [x for x in sdks if "iphoneos" in x]
        if not iphoneos:
            logger.error("No iphone SDK installed")
            ok = False
        else:
            iphoneos = iphoneos[0].split()[-1].replace("iphoneos", "")
            self.sdkver = iphoneos

        # get the latest iphonesimulator version
        iphonesim = [x for x in sdks if "iphonesimulator" in x]
        if not iphonesim:
            ok = False
            logger.error("Error: No iphonesimulator SDK installed")
        else:
            iphonesim = iphonesim[0].split()[-1].replace("iphonesimulator", "")
            self.sdksimver = iphonesim

        # get the path for Developer
        self.devroot = "{}/Platforms/iPhoneOS.platform/Developer".format(
            sh.xcode_select("-print-path").strip())

        # path to the iOS SDK
        self.iossdkroot = "{}/SDKs/iPhoneOS{}.sdk".format(
            self.devroot, self.sdkver)

        # root of the toolchain
        self.root_dir = realpath(dirname(__file__))
        self.build_dir = "{}/build".format(initial_working_directory)
        self.cache_dir = "{}/.cache".format(initial_working_directory)
        self.dist_dir = "{}/dist".format(initial_working_directory)
        self.install_dir = "{}/dist/root".format(initial_working_directory)
        self.include_dir = "{}/dist/include".format(initial_working_directory)
        self.archs = (
            Arch64Simulator(self),
            Arch64IOS(self))

        # path to some tools
        self.ccache = sh.which("ccache")
        for cython_fn in ("cython-2.7", "cython"):
            cython = sh.which(cython_fn)
            if cython:
                self.cython = cython
                break
        if not self.cython:
            ok = False
            logger.error("Missing requirement: cython is not installed")

        # check the basic tools
        for tool in ("pkg-config", "autoconf", "automake", "libtool"):
            if not sh.which(tool):
                logger.error("Missing requirement: {} is not installed".format(
                    tool))

        if not ok:
            sys.exit(1)

        self.use_pigz = sh.which('pigz')
        self.use_pbzip2 = sh.which('pbzip2')

        try:
            num_cores = int(sh.sysctl('-n', 'hw.ncpu'))
        except Exception:
            num_cores = None
        self.num_cores = num_cores if num_cores else 4  # default to 4 if we can't detect

        self.custom_recipes_paths = []
        ensure_dir(self.root_dir)
        ensure_dir(self.build_dir)
        ensure_dir(self.cache_dir)
        ensure_dir(self.dist_dir)
        ensure_dir(join(self.dist_dir, "frameworks"))
        ensure_dir(self.install_dir)
        ensure_dir(self.include_dir)
        ensure_dir(join(self.include_dir, "common"))

        # remove the most obvious flags that can break the compilation
        self.env.pop("MACOSX_DEPLOYMENT_TARGET", None)
        self.env.pop("PYTHONDONTWRITEBYTECODE", None)
        self.env.pop("ARCHFLAGS", None)
        self.env.pop("CFLAGS", None)
        self.env.pop("LDFLAGS", None)

        # set the state
        self.state = JsonStore(join(self.dist_dir, "state.db"))

    @property
    def concurrent_make(self):
        return "-j{}".format(self.num_cores)

    @property
    def concurrent_xcodebuild(self):
        return "IDEBuildOperationMaxNumberOfConcurrentCompileTasks={}".format(self.num_cores)


class Recipe:
    props = {
        "is_alias": False,
        "version": None,
        "url": None,
        "archs": [],
        "depends": [],
        "optional_depends": [],
        "python_depends": [],
        "library": None,
        "libraries": [],
        "include_dir": None,
        "include_per_arch": False,
        "include_name": None,
        "frameworks": [],
        "sources": [],
        "pbx_frameworks": [],
        "pbx_libraries": []
    }

    def __new__(cls):
        for prop, value in cls.props.items():
            if not hasattr(cls, prop):
                setattr(cls, prop, value)
        return super().__new__(cls)

    # API available for recipes
    def download_file(self, url, filename, cwd=None):
        """
        Download an `url` to `outfn`
        """
        if not url:
            return

        def report_hook(index, blksize, size):
            if size <= 0:
                progression = '{0} bytes'.format(index * blksize)
            else:
                progression = '{0:.2f}%'.format(
                        index * blksize * 100. / float(size))
            stdout.write('- Download {}\r'.format(progression))
            stdout.flush()

        if cwd:
            filename = join(cwd, filename)
        with suppress(FileNotFoundError):
            unlink(filename)

        # Clean up temporary files just in case before downloading.
        urlcleanup()

        logger.info('Downloading {0}'.format(url))
        attempts = 0
        while True:
            try:
                urlretrieve(url, filename, report_hook)
            except OSError:
                attempts += 1
                if attempts >= 5:
                    logger.error('Max download attempts reached: {}'.format(attempts))
                    raise
                logger.warning('Download failed. Retrying in 1 second...')
                time.sleep(1)
                continue
            break

        return filename

    def extract_file(self, filename, cwd):
        """
        Extract the `filename` into the directory `cwd`.
        """
        if not filename:
            return
        logger.info("Extract {} into {}".format(filename, cwd))
        if filename.endswith((".tgz", ".tar.gz")):
            if self.ctx.use_pigz:
                comp = '--use-compress-program={}'.format(self.ctx.use_pigz)
            else:
                comp = '-z'
            shprint(sh.tar, "-C", cwd, "-xv", comp, "-f", filename)

        elif filename.endswith((".tbz2", ".tar.bz2")):
            if self.ctx.use_pbzip2:
                comp = '--use-compress-program={}'.format(self.ctx.use_pbzip2)
            else:
                comp = '-j'
            shprint(sh.tar, "-C", cwd, "-xv", comp, "-f", filename)

        elif filename.endswith(".zip"):
            shprint(sh.unzip, "-d", cwd, filename)

        else:
            logger.error("Cannot extract, unrecognized extension for {}".format(
                filename))
            raise Exception()

    def get_archive_rootdir(self, filename):
        if filename.endswith((".tgz", ".tar.gz", ".tbz2", ".tar.bz2")):
            try:
                archive = tarfile.open(filename)
            except tarfile.ReadError:
                logger.warning('Error extracting the archive {0}'.format(filename))
                logger.warning(
                    'This is usually caused by a corrupt download. The file'
                    ' will be removed and re-downloaded on the next run.')
                logger.warning(filename)
                return

            root = archive.next().path.split("/")
            return root[0]
        elif filename.endswith(".zip"):
            with zipfile.ZipFile(filename) as zf:
                return dirname(zf.namelist()[0])
        else:
            logger.error("Unrecognized extension for {}."
                         " Cannot detect root directory".format(filename))
            raise Exception()

    def apply_patch(self, filename, target_dir=''):
        """
        Apply a patch from the current recipe directory into the current
        build directory.
        """
        target_dir = target_dir or self.build_dir
        logger.info("Apply patch {}".format(filename))
        filename = join(self.recipe_dir, filename)
        sh.patch("-t", "-d", target_dir, "-p1", "-i", filename)

    def copy_file(self, filename, dest):
        logger.info("Copy {} to {}".format(filename, dest))
        filename = join(self.recipe_dir, filename)
        dest = join(self.build_dir, dest)
        shutil.copy(filename, dest)

    def append_file(self, filename, dest):
        logger.info("Append {} to {}".format(filename, dest))
        filename = join(self.recipe_dir, filename)
        dest = join(self.build_dir, dest)
        with open(filename, "rb") as fd:
            data = fd.read()
        with open(dest, "ab") as fd:
            fd.write(data)

    def has_marker(self, marker):
        """
        Return True if the current build directory has the marker set
        """
        return exists(join(self.build_dir, ".{}".format(marker)))

    def set_marker(self, marker):
        """
        Set a marker info the current build directory
        """
        with open(join(self.build_dir, ".{}".format(marker)), "w") as fd:
            fd.write("ok")

    def delete_marker(self, marker):
        """
        Delete a specific marker
        """
        with suppress(FileNotFoundError):
            unlink(join(self.build_dir, ".{}".format(marker)))

    def get_include_dir(self):
        """
        Return the common include dir for this recipe
        """
        return join(self.ctx.include_dir, "common", self.name)

    def so_filename(self, name):
        """Return the filename of a library with the appropriate so suffix
        (.so for Python 2.7, .cpython-37m-darwin for Python 3.7)
        """
        return "{}{}".format(name, self.ctx.so_suffix)

    @property
    def name(self):
        modname = self.__class__.__module__
        return modname.split(".")[-1]

    @property
    def archive_fn(self):
        bfn = basename(self.url.format(version=self.version))
        fn = "{}/{}-{}".format(
            self.ctx.cache_dir,
            self.name, bfn)
        return fn

    @property
    def filtered_archs(self):
        result = []
        for arch in self.ctx.archs:
            if not self.archs or (arch.arch in self.archs):
                result.append(arch)
        return result

    @property
    def dist_libraries(self):
        libraries = []
        name = self.name
        if not name.startswith("lib"):
            name = "lib{}".format(name)
        if self.library:
            static_fn = join(self.ctx.dist_dir, "lib", "{}.a".format(name))
            libraries.append(static_fn)
        for library in self.libraries:
            static_fn = join(self.ctx.dist_dir, "lib", basename(library))
            libraries.append(static_fn)
        return libraries

    def get_build_dir(self, arch):
        return join(self.ctx.build_dir, self.name, arch, self.archive_root)

    # Public Recipe API to be subclassed if needed

    def init_with_ctx(self, ctx):
        self.ctx = ctx
        include_dir = None
        if self.include_dir:
            include_name = self.include_name or self.name
            if self.include_per_arch:
                include_dir = join("{arch.arch}", include_name)
            else:
                include_dir = join("common", include_name)
        if include_dir:
            logger.info("Include dir added: {}".format(include_dir))
            self.ctx.include_dirs.append(include_dir)

    def get_recipe_env(self, arch=None):
        """Return the env specialized for the recipe
        """
        if arch is None:
            arch = self.filtered_archs[0]
        return arch.get_env()

    def set_hostpython(self, instance, version):
        state = self.ctx.state
        hostpython = state.get("hostpython")
        if hostpython is None:
            state["hostpython"] = instance.name
            state.sync()
        elif hostpython != instance.name:
            logger.error("Wanted to use {}".format(instance.name))
            logger.error("but hostpython is already provided by {}.".format(
                hostpython))
            logger.error("You can have only one hostpython version compiled")
            sys.exit(1)
        self.ctx.hostpython_ver = version
        self.ctx.hostpython_recipe = instance

    def set_python(self, instance, version):
        state = self.ctx.state
        python = state.get("python")
        if python is None:
            state["python"] = instance.name
            state.sync()
        elif python != instance.name:
            logger.error("Wanted to use {}".format(instance.name))
            logger.error("but python is already provided by {}.".format(
                python))
            logger.error("You can have only one python version compiled")
            sys.exit(1)
        self.ctx.python_ver = version
        self.ctx.python_recipe = instance

    @property
    def archive_root(self):
        key = "{}.archive_root".format(self.name)
        value = self.ctx.state.get(key)
        if not value and self.url != "":
            value = self.get_archive_rootdir(self.archive_fn)
            if value is not None:
                self.ctx.state[key] = value
        return value

    def execute(self):
        if self.custom_dir:
            self.ctx.state.remove_all(self.name)
        self.download()
        self.extract()
        self.build_all()

    @property
    def custom_dir(self):
        """Check if there is a variable name to specify a custom version /
        directory to use instead of the current url.
        """
        envname = "{}_DIR".format(self.name.upper())
        d = environ.get(envname)
        if not d:
            return
        if not exists(d):
            raise ValueError("Invalid path passed into {}".format(envname))
        return d

    def init_after_import(cls, ctx):
        """This can be used to dynamically set some variables
        depending of the state
        """
        pass

    @cache_execution
    def download(self):
        key = "{}.archive_root".format(self.name)
        if self.custom_dir:
            self.ctx.state[key] = basename(self.custom_dir)
        else:
            src_dir = join(self.recipe_dir, self.url)
            if exists(src_dir):
                self.ctx.state[key] = basename(src_dir)
                return
            fn = self.archive_fn
            if not exists(fn):
                self.download_file(self.url.format(version=self.version), fn)
            status = self.get_archive_rootdir(self.archive_fn)
            if status is not None:
                self.ctx.state[key] = status

    @cache_execution
    def extract(self):
        # recipe tmp directory
        for arch in self.filtered_archs:
            logger.info("Extract {} for {}".format(self.name, arch.arch))
            self.extract_arch(arch.arch)

    def extract_arch(self, arch):
        build_dir = join(self.ctx.build_dir, self.name, arch)
        dest_dir = join(build_dir, self.archive_root)
        if self.custom_dir:
            shutil.rmtree(dest_dir, ignore_errors=True)
            shutil.copytree(self.custom_dir, dest_dir)
        else:
            if exists(dest_dir):
                return
            src_dir = join(self.recipe_dir, self.url)
            if exists(src_dir):
                shutil.copytree(src_dir, dest_dir)
                return
            ensure_dir(build_dir)
            self.extract_file(self.archive_fn, build_dir)

    @cache_execution
    def build(self, arch):
        self.build_dir = self.get_build_dir(arch.arch)
        if self.has_marker("building"):
            logger.warning("{} build for {} has been incomplete".format(
                self.name, arch.arch))
            logger.warning("Warning: deleting the build and restarting.")
            shutil.rmtree(self.build_dir, ignore_errors=True)
            self.extract_arch(arch.arch)

        if self.has_marker("build_done"):
            logger.info("Build python for {} already done.".format(arch.arch))
            return

        self.set_marker("building")

        chdir(self.build_dir)
        logger.info("Prebuild {} for {}".format(self.name, arch.arch))
        self.prebuild_arch(arch)
        logger.info("Build {} for {}".format(self.name, arch.arch))
        self.build_arch(arch)
        logger.info("Postbuild {} for {}".format(self.name, arch.arch))
        self.postbuild_arch(arch)
        self.delete_marker("building")
        self.set_marker("build_done")

    @cache_execution
    def build_all(self):
        filtered_archs = self.filtered_archs
        logger.info("Build {} for {} (filtered)".format(
            self.name,
            ", ".join([x.arch for x in filtered_archs])))
        for arch in self.filtered_archs:
            self.build(arch)

        name = self.name
        if self.library:
            logger.info("Create lipo library for {}".format(name))
            if not name.startswith("lib"):
                name = "lib{}".format(name)
            static_fn = join(self.ctx.dist_dir, "lib", "{}.a".format(name))
            ensure_dir(dirname(static_fn))
            logger.info("Lipo {} to {}".format(self.name, static_fn))
            self.make_lipo(static_fn)
        if self.libraries:
            logger.info("Create multiple lipo for {}".format(name))
            for library in self.libraries:
                static_fn = join(self.ctx.dist_dir, "lib", basename(library))
                ensure_dir(dirname(static_fn))
                logger.info("  - Lipo-ize {}".format(library))
                self.make_lipo(static_fn, library)
        logger.info("Install include files for {}".format(self.name))
        self.install_include()
        logger.info("Install frameworks for {}".format(self.name))
        self.install_frameworks()
        logger.info("Install sources for {}".format(self.name))
        self.install_sources()
        logger.info("Install python deps for {}".format(self.name))
        self.install_python_deps()
        logger.info("Install {}".format(self.name))
        self.install()

    def prebuild_arch(self, arch):
        prebuild = "prebuild_{}".format(arch.arch)
        logger.debug("Invoking {}".format(prebuild))
        if hasattr(self, prebuild):
            getattr(self, prebuild)()

    def build_arch(self, arch):
        build = "build_{}".format(arch.arch)
        logger.debug("Invoking {}".format(build))
        if hasattr(self, build):
            getattr(self, build)()

    def postbuild_arch(self, arch):
        postbuild = "postbuild_{}".format(arch.arch)
        logger.debug("Invoking {}".format(postbuild))
        if hasattr(self, postbuild):
            getattr(self, postbuild)()
        remove_junk(self.build_dir)

    def update_state(self, key, value):
        """
        Update entry in state database. This is usually done in the
        @cache_execution decorator to log an action and its time of occurrence,
        but it needs to be done manually in recipes.
        """
        key_time = "{}.at".format(key)
        self.ctx.state[key] = value
        now_str = str(datetime.utcnow())
        self.ctx.state[key_time] = now_str
        logger.debug("New State: {} at {}".format(key, now_str))

    @cache_execution
    def make_lipo(self, filename, library=None):
        if library is None:
            library = self.library
        if not library:
            return
        args = []
        for arch in self.filtered_archs:
            library_fn = library.format(arch=arch)
            args += [
                "-arch", arch.arch,
                join(self.get_build_dir(arch.arch), library_fn)]
        shprint(sh.lipo, "-create", "-output", filename, *args)

    @cache_execution
    def install_frameworks(self):
        if not self.frameworks:
            return
        arch = self.filtered_archs[0]
        build_dir = self.get_build_dir(arch.arch)
        for framework in self.frameworks:
            logger.info("Install Framework {}".format(framework))
            src = join(build_dir, framework)
            dest = join(self.ctx.dist_dir, "frameworks", framework)
            ensure_dir(dirname(dest))
            shutil.rmtree(dest, ignore_errors=True)
            logger.debug("Copy {} to {}".format(src, dest))
            shutil.copytree(src, dest)

    @cache_execution
    def install_sources(self):
        if not self.sources:
            return
        arch = self.filtered_archs[0]
        build_dir = self.get_build_dir(arch.arch)
        for source in self.sources:
            logger.info("Install Sources{}".format(source))
            src = join(build_dir, source)
            dest = join(self.ctx.dist_dir, "sources", self.name)
            ensure_dir(dirname(dest))
            shutil.rmtree(dest, ignore_errors=True)
            logger.debug("Copy {} to {}".format(src, dest))
            shutil.copytree(src, dest)

    @cache_execution
    def install_include(self):
        if not self.include_dir:
            return
        if self.include_per_arch:
            archs = self.ctx.archs
        else:
            archs = self.filtered_archs[:1]

        include_dirs = self.include_dir
        if not isinstance(include_dirs, (list, tuple)):
            include_dirs = list([include_dirs])

        for arch in archs:
            arch_dir = "common"
            if self.include_per_arch:
                arch_dir = arch.arch
            include_name = self.include_name or self.name
            dest_dir = join(self.ctx.include_dir, arch_dir, include_name)
            shutil.rmtree(dest_dir, ignore_errors=True)
            build_dir = self.get_build_dir(arch.arch)

            for include_dir in include_dirs:
                dest_name = None
                if isinstance(include_dir, (list, tuple)):
                    include_dir, dest_name = include_dir
                include_dir = include_dir.format(arch=arch, ctx=self.ctx)
                src_dir = join(build_dir, include_dir)
                if dest_name is None:
                    dest_name = basename(src_dir)
                if isdir(src_dir):
                    shutil.copytree(src_dir, dest_dir)
                else:
                    dest = join(dest_dir, dest_name)
                    logger.info("Copy Include {} to {}".format(src_dir, dest))
                    ensure_dir(dirname(dest))
                    shutil.copy(src_dir, dest)

    @cache_execution
    def install_python_deps(self):
        for dep in self.python_depends:
            _pip(["install", dep])

    @cache_execution
    def install(self):
        pass

    @classmethod
    def list_recipes(cls, **kwargs):
        skip_list = kwargs.pop("skip_list", ['__pycache__'])
        recipes_dir = join(dirname(__file__), "recipes")
        for name in sorted(listdir(recipes_dir)):
            fn = join(recipes_dir, name)
            if isdir(fn) and name not in skip_list:
                yield name

    @classmethod
    def get_recipe(cls, name, ctx):
        if not hasattr(cls, "recipes"):
            cls.recipes = {}

        if '==' in name:
            name, version = name.split('==')
        else:
            version = None

        if name in cls.recipes:
            recipe = cls.recipes[name]
        else:
            try:
                mod = importlib.import_module("kivy_ios.recipes.{}".format(name))
                recipe = mod.recipe
                recipe.recipe_dir = join(ctx.root_dir, "recipes", name)
            except ImportError:
                logger.info("Looking for recipe '{}' in custom recipes paths (if provided)".format(name))
                for custom_recipe_path in ctx.custom_recipes_paths:
                    if custom_recipe_path.split('/')[-1] == name:
                        spec = importlib.util.spec_from_file_location(name,
                                                                      join(custom_recipe_path, '__init__.py'))
                        mod = importlib.util.module_from_spec(spec)
                        spec.loader.exec_module(mod)
                        recipe = mod.recipe
                        recipe.recipe_dir = custom_recipe_path
                        logger.info("Custom recipe '{}' found in folder {}".format(name, custom_recipe_path))
                        break
                else:
                    raise
            recipe.init_after_import(ctx)

        if version:
            recipe.version = version

        return recipe


class PythonRecipe(Recipe):
    @cache_execution
    def install(self):
        self.install_python_package()
        self.reduce_python_package()
        remove_junk(self.ctx.site_packages_dir)

    def install_python_package(self, name=None, env=None, is_dir=True):
        """Automate the installation of a Python package into the target
        site-packages.

        It will works with the first filtered_archs, and the name of the recipe.
        """
        arch = self.filtered_archs[0]
        if name is None:
            name = self.name
        if env is None:
            env = self.get_recipe_env(arch)
        logger.info("Install {} into the site-packages".format(name))
        build_dir = self.get_build_dir(arch.arch)
        chdir(build_dir)
        hostpython = sh.Command(self.ctx.hostpython)

        shprint(
            hostpython,
            "setup.py",
            "install",
            "-O2",
            "--root", self.ctx.python_prefix,
            "--prefix", "",
            _env=env,
        )

    def reduce_python_package(self):
        """Feel free to remove things you don't want in the final
        site-packages.
        """
        pass


class CythonRecipe(PythonRecipe):
    pre_build_ext = False
    cythonize = True

    def cythonize_file(self, filename):
        if filename.startswith(self.build_dir):
            filename = filename[len(self.build_dir) + 1:]
        logger.info("Cythonize {}".format(filename))
        # note when kivy-ios package installed the `cythonize.py` script
        # doesn't (yet) have the executable bit hence we explicitly call it
        # with the Python interpreter
        cythonize_script = join(self.ctx.root_dir, "tools", "cythonize.py")
        shprint(sh.python, cythonize_script, filename)

    def cythonize_build(self):
        if not self.cythonize:
            return
        root_dir = self.build_dir
        for root, dirnames, filenames in walk(root_dir):
            for filename in fnmatch.filter(filenames, "*.pyx"):
                self.cythonize_file(join(root, filename))

    def biglink(self):
        dirs = []
        for root, dirnames, filenames in walk(self.build_dir):
            if fnmatch.filter(filenames, "*.so.libs"):
                dirs.append(root)
        cmd = sh.Command(join(self.ctx.root_dir, "tools", "biglink"))
        shprint(cmd, join(self.build_dir, "lib{}.a".format(self.name)), *dirs)

    def get_recipe_env(self, arch):
        env = super().get_recipe_env(arch)
        env["KIVYIOSROOT"] = self.ctx.root_dir
        env["IOSSDKROOT"] = arch.sysroot
        env["CUSTOMIZED_OSX_COMPILER"] = 'True'
        env["LDSHARED"] = join(self.ctx.root_dir, "tools", "liblink")
        env["ARM_LD"] = env["LD"]
        env["ARCH"] = arch.arch
        return env

    def build_arch(self, arch):
        build_env = self.get_recipe_env(arch)
        hostpython = sh.Command(self.ctx.hostpython)
        if self.pre_build_ext:
            with suppress(Exception):
                shprint(hostpython, "setup.py", "build_ext", "-g",
                        _env=build_env)
        self.cythonize_build()
        shprint(hostpython, "setup.py", "build_ext", "-g",
                _env=build_env)
        self.biglink()


def build_recipes(names, ctx):
    # gather all the dependencies
    logger.info("Want to build {}".format(names))
    graph = Graph()
    ctx.wanted_recipes = names[:]
    recipe_to_load = names
    recipe_loaded = []
    while names:
        name = recipe_to_load.pop(0)
        if name in recipe_loaded:
            continue
        try:
            recipe = Recipe.get_recipe(name, ctx)
        except KeyError:
            logger.error("No recipe named {}".format(name))
            sys.exit(1)
        graph.add(name, name)
        logger.info("Loaded recipe {} (depends of {}, optional are {})".format(
            name, recipe.depends, recipe.optional_depends))
        for depend in recipe.depends:
            graph.add(name, depend)
            recipe_to_load += recipe.depends
        for depend in recipe.optional_depends:
            # in case of compilation after the initial one, take in account
            # of the already compiled recipes
            key = "{}.build_all".format(depend)
            if key in ctx.state:
                recipe_to_load.append(name)
                graph.add(name, depend)
            else:
                graph.add_optional(name, depend)
        recipe_loaded.append(name)

    build_order = list(graph.find_order())
    logger.info("Build order is {}".format(build_order))
    recipes = [Recipe.get_recipe(name, ctx) for name in build_order]
    recipes = [recipe for recipe in recipes if not recipe.is_alias]
    recipes_order = [recipe.name for recipe in recipes]
    logger.info("Recipe order is {}".format(recipes_order))
    for recipe in recipes:
        recipe.init_with_ctx(ctx)
    for recipe in recipes:
        recipe.execute()


def ensure_dir(filename):
    makedirs(filename, exist_ok=True)


def ensure_recipes_loaded(ctx):
    for recipe in Recipe.list_recipes():
        key = "{}.build_all".format(recipe)
        if key not in ctx.state:
            continue
        recipe = Recipe.get_recipe(recipe, ctx)
        recipe.init_with_ctx(ctx)


def _pip(args):
    ctx = Context()
    for recipe in Recipe.list_recipes():
        key = "{}.build_all".format(recipe)
        if key not in ctx.state:
            continue
        recipe = Recipe.get_recipe(recipe, ctx)
        recipe.init_with_ctx(ctx)
    if not hasattr(ctx, "site_packages_dir"):
        logger.error("python must be compiled before using pip")
        sys.exit(1)

    pip_env = {
        "CC": "/bin/false",
        "CXX": "/bin/false",
        "PYTHONPATH": ctx.site_packages_dir,
        "PYTHONOPTIMIZE": "2",
        # "PIP_INSTALL_TARGET": ctx.site_packages_dir
    }

    pip_path = join(ctx.dist_dir, 'hostpython3', 'bin', 'pip3')

    if len(args) > 1 and args[0] == "install":
        pip_args = ["--isolated", "--prefix", ctx.python_prefix]
        args = ["install"] + pip_args + args[1:]

    logger.error("Executing pip with: {}".format(args))
    pip_cmd = sh.Command(pip_path)
    shprint(pip_cmd, *args, _env=pip_env)


def update_pbxproj(filename, pbx_frameworks=None):
    # list all the compiled recipes
    ctx = Context()
    pbx_libraries = []
    if pbx_frameworks is None:
        pbx_frameworks = []
    frameworks = []
    libraries = []
    sources = []
    for recipe in Recipe.list_recipes():
        key = "{}.build_all".format(recipe)
        if key not in ctx.state:
            continue
        recipe = Recipe.get_recipe(recipe, ctx)
        recipe.init_with_ctx(ctx)
        pbx_frameworks.extend(recipe.pbx_frameworks)
        pbx_libraries.extend(recipe.pbx_libraries)
        libraries.extend(recipe.dist_libraries)
        frameworks.extend(recipe.frameworks)
        if recipe.sources:
            sources.append(recipe.name)

    pbx_frameworks = list(set(pbx_frameworks))
    pbx_libraries = list(set(pbx_libraries))
    libraries = list(set(libraries))

    logger.info("-" * 70)
    logger.info("The project need to have:")
    logger.info("iOS Frameworks: {}".format(pbx_frameworks))
    logger.info("iOS Libraries: {}".format(pbx_libraries))
    logger.info("iOS local Frameworks: {}".format(frameworks))
    logger.info("Libraries: {}".format(libraries))
    logger.info("Sources to link: {}".format(sources))

    logger.info("-" * 70)
    logger.info("Analysis of {}".format(filename))

    project = XcodeProject.load(filename)
    sysroot = sh.xcrun("--sdk", "iphonesimulator", "--show-sdk-path").strip()

    group = project.get_or_create_group("Frameworks")
    g_classes = project.get_or_create_group("Classes")
    file_options = FileOptions(embed_framework=False, code_sign_on_copy=True)
    for framework in pbx_frameworks:
        framework_name = "{}.framework".format(framework)
        if framework_name in frameworks:
            logger.info("Ensure {} is in the project (pbx_frameworks, local)".format(framework))
            f_path = join(ctx.dist_dir, "frameworks", framework_name)
        else:
            logger.info("Ensure {} is in the project (pbx_frameworks, system)".format(framework))
            f_path = join(sysroot, "System", "Library", "Frameworks",
                          "{}.framework".format(framework))
        project.add_file(f_path, parent=group, tree="DEVELOPER_DIR",
                         force=False, file_options=file_options)
    for library in pbx_libraries:
        logger.info("Ensure {} is in the project (pbx_libraries, dylib+tbd)".format(library))
        f_path = join(sysroot, "usr", "lib",
                      "{}.dylib".format(library))
        project.add_file(f_path, parent=group, tree="DEVELOPER_DIR", force=False)
        f_path = join(sysroot, "usr", "lib",
                      "{}.tbd".format(library))
        project.add_file(f_path, parent=group, tree="DEVELOPER_DIR", force=False)
    for library in libraries:
        logger.info("Ensure {} is in the project (libraries)".format(library))
        project.add_file(library, parent=group, force=False)
    for name in sources:
        logger.info("Ensure {} sources are used".format(name))
        fn = join(ctx.dist_dir, "sources", name)
        project.add_folder(fn, parent=g_classes)

    project.backup()
    project.save()


class ToolchainCL:
    def __init__(self):
        parser = argparse.ArgumentParser(
                description="Tool for managing the iOS / Python toolchain",
                usage="""toolchain <command> [<args>]

Available commands:
build         Build a recipe (compile a library for the required target
              architecture)
clean         Clean the build of the specified recipe
distclean     Clean the build and the result
recipes       List all the available recipes
status        List all the recipes and their build status
build_info    Display the current build context and Architecture info

Xcode:
create        Create a new xcode project
update        Update an existing xcode project (frameworks, libraries..)
launchimage   Create Launch images for your xcode project
icon          Create Icons for your xcode project
pip           Install a pip dependency into the distribution
""")
        parser.add_argument("command", help="Command to run")
        args = parser.parse_args(sys.argv[1:2])
        if not hasattr(self, args.command):
            print('Unrecognized command')
            parser.print_help()
            exit(1)
        getattr(self, args.command)()

    @staticmethod
    def find_xcodeproj(filename):
        if not filename.endswith(".xcodeproj"):
            # try to find the xcodeproj
            from glob import glob
            xcodeproj = glob(join(filename, "*.xcodeproj"))
            if not xcodeproj:
                logger.error("Unable to find a xcodeproj in {}".format(filename))
                sys.exit(1)
            filename = xcodeproj[0]
        return filename

    def build(self):
        ctx = Context()
        parser = argparse.ArgumentParser(
                description="Build the toolchain")
        parser.add_argument("recipe", nargs="+", help="Recipe to compile")
        parser.add_argument("--arch", action="append",
                            help="Restrict compilation to this arch")
        parser.add_argument("--concurrency", type=int, default=ctx.num_cores,
                            help="number of concurrent build processes (where supported)")
        parser.add_argument("--no-pigz", action="store_true", default=not bool(ctx.use_pigz),
                            help="do not use pigz for gzip decompression")
        parser.add_argument("--no-pbzip2", action="store_true", default=not bool(ctx.use_pbzip2),
                            help="do not use pbzip2 for bzip2 decompression")
        parser.add_argument("--add-custom-recipe", action="append", default=[],
                            help="Path to custom recipe")
        args = parser.parse_args(sys.argv[2:])

        if args.arch:
            if len(args.arch) == 1:
                archs = args.arch[0].split()
            else:
                archs = args.arch
            available_archs = [arch.arch for arch in ctx.archs]
            for arch in archs[:]:
                if arch not in available_archs:
                    logger.error("Architecture {} invalid".format(arch))
                    archs.remove(arch)
                    continue
            ctx.archs = [arch for arch in ctx.archs if arch.arch in archs]
            logger.info("Architectures restricted to: {}".format(archs))
        ctx.num_cores = args.concurrency
        if args.no_pigz:
            ctx.use_pigz = False
        if args.no_pbzip2:
            ctx.use_pbzip2 = False
        ctx.use_pigz = ctx.use_pbzip2
        logger.info("Building with {} processes, where supported".format(ctx.num_cores))
        if ctx.use_pigz:
            logger.info("Using pigz to decompress gzip data")
        if ctx.use_pbzip2:
            logger.info("Using pbzip2 to decompress bzip2 data")
        for custom_recipe_path in args.add_custom_recipe:
            if exists(custom_recipe_path):
                logger.info(f"Adding {custom_recipe_path} to custom recipes paths")
                ctx.custom_recipes_paths.append(custom_recipe_path)
            else:
                logger.error(f"{custom_recipe_path} isn't a valid path")
        build_recipes(args.recipe, ctx)

    def recipes(self):
        parser = argparse.ArgumentParser(
                description="List all the available recipes")
        parser.add_argument(
                "--compact", action="store_true",
                help="Produce a compact list suitable for scripting")
        args = parser.parse_args(sys.argv[2:])

        if args.compact:
            print(" ".join(list(Recipe.list_recipes())))
        else:
            ctx = Context()
            for name in Recipe.list_recipes():
                with suppress(Exception):
                    recipe = Recipe.get_recipe(name, ctx)
                    print("{recipe.name:<12} {recipe.version:<8}".format(recipe=recipe))

    def clean(self):
        def clean_cache(recipe, ctx):
            """ Remove download artifacts for this build. """
            recipe_inst = Recipe.get_recipe(recipe, ctx)
            recipe_inst.ctx = ctx
            if exists(recipe_inst.archive_fn):
                unlink(recipe_inst.archive_fn)

        parser = argparse.ArgumentParser(
                description="Clean the build")
        parser.add_argument("recipe", nargs="*", help="Recipe to clean")
        args = parser.parse_args(sys.argv[2:])
        ctx = Context()
        if args.recipe:
            for recipe in args.recipe:
                logger.info("Cleaning {} build".format(recipe))
                ctx.state.remove_all("{}.".format(recipe))
                build_dir = join(ctx.build_dir, recipe)
                shutil.rmtree(build_dir, ignore_errors=True)
                clean_cache(recipe, ctx)
        else:
            logger.info("Delete build directory")
            shutil.rmtree(ctx.build_dir, ignore_errors=True)

    def distclean(self):
        ctx = Context()
        shutil.rmtree(ctx.build_dir, ignore_errors=True)
        shutil.rmtree(ctx.dist_dir, ignore_errors=True)
        shutil.rmtree(ctx.cache_dir, ignore_errors=True)

    def status(self):
        ctx = Context()
        for recipe in Recipe.list_recipes():
            key = "{}.build_all".format(recipe)
            keytime = "{}.build_all.at".format(recipe)

            if key in ctx.state:
                status = "Build OK (built at {})".format(ctx.state[keytime])
            else:
                status = "Not built"
            print("{:<12} - {}".format(
                recipe, status))

    def create(self):
        parser = argparse.ArgumentParser(
                description="Create a new xcode project")
        parser.add_argument("name", help="Name of your project")
        parser.add_argument("directory", help="Directory where your project lives")
        parser.add_argument("--add-framework", action="append", help="Additional Frameworks to include with this project")
        args = parser.parse_args(sys.argv[2:])

        from cookiecutter.main import cookiecutter
        ctx = Context()
        ensure_recipes_loaded(ctx)

        if not hasattr(ctx, "python_ver"):
            logger.error("No python recipe compiled!")
            logger.error("You must have compiled at least python3")
            logger.error("recipe to be able to create a project.")
            sys.exit(1)

        template_dir = join(curdir, "tools", "templates")
        context = {
            "title": args.name,
            "project_name": args.name.lower(),
            "domain_name": "org.kivy.{}".format(args.name.lower()),
            "project_dir": realpath(args.directory),
            "version": "1.0.0",
            "dist_dir": ctx.dist_dir,
        }
        cookiecutter(template_dir, no_input=True, extra_context=context)
        filename = join(
                initial_working_directory,
                "{}-ios".format(args.name.lower()),
                "{}.xcodeproj".format(args.name.lower()),
                "project.pbxproj")
        update_pbxproj(filename, pbx_frameworks=args.add_framework)
        print("--")
        print("Project directory : {}-ios".format(
            args.name.lower()))
        print("XCode project     : {0}-ios/{0}.xcodeproj".format(
            args.name.lower()))

    def update(self):
        parser = argparse.ArgumentParser(
                description="Update an existing xcode project")
        parser.add_argument("filename", help="Path to your project or xcodeproj")
        parser.add_argument("--add-framework", action="append", help="Additional Frameworks to include with this project")
        args = parser.parse_args(sys.argv[2:])

        filename = self.find_xcodeproj(args.filename)
        filename = join(filename, "project.pbxproj")
        if not exists(filename):
            logger.error("{} not found".format(filename))
            sys.exit(1)

        update_pbxproj(filename, pbx_frameworks=args.add_framework)
        print("--")
        print("Project {} updated".format(filename))

    def build_info(self):
        ctx = Context()
        print("Build Context")
        print("-------------")
        for attr in dir(ctx):
            if not attr.startswith("_"):
                if not callable(attr) and attr != 'archs':
                    print("{}: {}".format(attr, pformat(getattr(ctx, attr))))
        for arch in ctx.archs:
            ul = '-' * (len(str(arch))+6)
            print("\narch: {}\n{}".format(str(arch), ul))
            for attr in dir(arch):
                if not attr.startswith("_"):
                    if not callable(attr) and attr not in ['arch', 'ctx', 'get_env']:
                        print("{}: {}".format(attr, pformat(getattr(arch, attr))))
            env = arch.get_env()
            print("env ({}): {}".format(arch, pformat(env)))

    def pip3(self):
        self.pip()

    def pip(self):
        _pip(sys.argv[2:])

    def launchimage(self):
        import xcassets
        self._xcassets("LaunchImage", xcassets.launchimage)

    def icon(self):
        import xcassets
        self._xcassets("Icon", xcassets.icon)

    def xcode(self):
        parser = argparse.ArgumentParser(description="Open the xcode project")
        parser.add_argument("filename", help="Path to your project or xcodeproj")
        args = parser.parse_args(sys.argv[2:])
        filename = self.find_xcodeproj(args.filename)
        sh.open(filename)

    def _xcassets(self, title, command):
        parser = argparse.ArgumentParser(
                description="Generate {} for your project".format(title))
        parser.add_argument("filename", help="Path to your project or xcodeproj")
        parser.add_argument("image", help="Path to your initial {}.png".format(title.lower()))
        args = parser.parse_args(sys.argv[2:])

        if not exists(args.image):
            logger.error("image path does not exists.")
            return

        filename = self.find_xcodeproj(args.filename)
        project_name = filename.split("/")[-1].replace(".xcodeproj", "")
        images_xcassets = realpath(join(filename, "..", project_name,
                                        "Images.xcassets"))
        if not exists(images_xcassets):
            logger.warning("Images.xcassets not found, creating it.")
            makedirs(images_xcassets)
        logger.info("Images.xcassets located at {}".format(images_xcassets))

        command(images_xcassets, args.image)


def main():
    ToolchainCL()


if __name__ == "__main__":
    main()



from kivy_ios.toolchain import CythonRecipe


class CurlyRecipe(CythonRecipe):
    version = "master"
    url = "https://github.com/tito/curly/archive/{version}.zip"
    library = "libcurly.a"
    depends = ["python", "libcurl", "sdl2", "sdl2_image"]
    pre_build_ext = True


recipe = CurlyRecipe()



"""
Author: Lawrence Du
E-mail: larrydu88@gmail.com
"""

from kivy_ios.toolchain import CythonRecipe, shprint
import sh
from os.path import join
from os import chdir
import logging

logger = logging.getLogger(__name__)


class KiventCoreRecipe(CythonRecipe):
    version = 'master'
    url = 'https://github.com/kivy/kivent/archive/{version}.zip'
    name = 'kivent_core'
    depends = ['libffi', 'kivy']  # note: unsure if libffi is necessary here
    pre_build_ext = False
    subbuilddir = False
    cythonize = True
    pbx_frameworks = ["OpenGLES"]  # note: This line may be unnecessary

    def get_recipe_env(self, arch):
        env = super(KiventCoreRecipe, self).get_recipe_env(arch)
        env['CYTHONPATH'] = self.get_recipe(
            'kivy', self.ctx).get_build_dir(arch.arch)
        return env

    def get_build_dir(self, arch, sub=False):
        """
        Call this to get the correct build_dir, where setup.py is located which is
        actually under modules/core/setup.py
        """
        builddir = super(KiventCoreRecipe, self).get_build_dir(str(arch))
        if sub or self.subbuilddir:
            core_build_dir = join(builddir, 'modules', 'core')
            logger.info("Core build directory is located at {}".format(core_build_dir))
            return core_build_dir
        else:
            logger.info("Building in {}".format(builddir))
            return builddir

    def build_arch(self, arch):
        """
        Override build.arch to avoid calling setup.py here (Call it in
        install() instead).
        """
        self.subbuildir = True
        self.cythonize_build()
        self.biglink()
        self.subbuilddir = False

    def install(self):
        """
        This method simply builds the command line call for calling
        kivent_core/modules/core/setup.py

        This constructs the equivalent of the command
        "$python setup.py build_ext install"
        only with the environment variables altered for each different architecture
        The appropriate version of kivy also needs to be added to the path, and this
        differs for each architecture (i386, x86_64, armv7, etc)

        Note: This method is called by build_all() in toolchain.py

        """
        arch = list(self.filtered_archs)[0]

        build_dir = self.get_build_dir(arch.arch, sub=True)
        logger.info("Building kivent_core {} in {}".format(arch.arch, build_dir))
        chdir(build_dir)
        hostpython = sh.Command(self.ctx.hostpython)

        # Get the appropriate environment for this recipe (including CYTHONPATH)
        # build_env = arch.get_env()
        build_env = self.get_recipe_env(arch)

        dest_dir = join(self.ctx.dist_dir, "root", "python")
        build_env['PYTHONPATH'] = join(dest_dir, 'lib', 'python3.7', 'site-packages')

        # Add Architecture specific kivy path for 'import kivy' to PYTHONPATH
        arch_kivy_path = self.get_recipe('kivy', self.ctx).get_build_dir(arch.arch)
        build_env['PYTHONPATH'] = join(build_env['PYTHONPATH'], ':', arch_kivy_path)

        # Make sure you call kivent_core/modules/core/setup.py
        subdir_path = self.get_build_dir(str(arch), sub=True)
        setup_path = join(subdir_path, "setup.py")

        # Print out directories for sanity check
        logger.info("ENVS", build_env)
        logger.info("ROOT", self.ctx.root_dir)
        logger.info("BUILD", self.ctx.build_dir)
        logger.info("INCLUDE", self.ctx.include_dir)
        logger.info("DISTDIR", self.ctx.dist_dir)
        logger.info("ARCH KIVY LOC", self.get_recipe('kivy', self.ctx).get_build_dir(arch.arch))

        shprint(hostpython, setup_path, "build_ext", "install", _env=build_env)


recipe = KiventCoreRecipe()



from kivy_ios.toolchain import Recipe, shprint
from os.path import join
import sh


class JpegRecipe(Recipe):
    version = "v9a"
    url = "http://www.ijg.org/files/jpegsrc.{version}.tar.gz"
    library = ".libs/libjpeg.a"
    include_dir = [
        ("jpeglib.h", ""),
        ("jconfig.h", ""),
        ("jerror.h", ""),
        ("jmorecfg.h", ""),
        ]
    include_per_arch = True

    def build_arch(self, arch):
        build_env = arch.get_env()
        configure = sh.Command(join(self.build_dir, "configure"))
        shprint(configure,
                "CC={}".format(build_env["CC"]),
                "LD={}".format(build_env["LD"]),
                "CFLAGS={}".format(build_env["CFLAGS"]),
                "LDFLAGS={}".format(build_env["LDFLAGS"]),
                "--prefix=/",
                "--host={}".format(arch.triple),
                "--disable-shared")
        shprint(sh.make, "clean")
        shprint(sh.make, self.ctx.concurrent_make)


recipe = JpegRecipe()



"""
Author: Lawrence Du, Lukasz Mach
E-mail: larrydu88@gmail.com, maho@pagema.net
"""

from kivy_ios.toolchain import CythonRecipe


class CymunkRecipe(CythonRecipe):
    version = 'master'
    url = 'https://github.com/kivy/cymunk/archive/{version}.zip'
    name = 'cymunk'
    pre_build_ext = True
    library = 'libcymunk.a'

    depends = ['python']

    def get_recipe_env(self, arch):
        ret = super(CymunkRecipe, self).get_recipe_env(arch)
        ret['CFLAGS'] += ' -Wno-implicit-function-declaration'
        return ret


recipe = CymunkRecipe()



# -*- coding: utf-8 -*-
import sys
from kivy_ios.toolchain import Recipe
import logging

logger = logging.getLogger(__name__)


class HostpythonAliasRecipe(Recipe):
    """
    Note this recipe was created to handle both hostpython2 and hostpython3.
    As hostpython2 support was dropped, this could probably be simplified.
    """
    is_alias = True

    def init_after_import(self, ctx):
        hostpython = ctx.state.get("hostpython")
        if not hostpython:
            # search in wanted_recipes if it's the first time
            if "hostpython3" in ctx.wanted_recipes:
                hostpython = "hostpython3"
            else:
                logger.error("No hostpython version set in the build.")
                logger.error("Add python3 in your recipes:")
                logger.error("./toolchain.py build python3 ...")
                sys.exit(1)
        if hostpython:
            self.depends = [hostpython]


recipe = HostpythonAliasRecipe()



'''
IOS module
==========

IOS module is wrapping some part of the IOS features.

'''

from cpython cimport Py_INCREF, Py_DECREF
from os.path import basename

cdef extern from "ios_wrapper.h":
    ctypedef void (*ios_send_email_cb)(char *, void *)
    ctypedef struct padding:
        float top
        float bottom
        float right
        float left
    int ios_send_email(char *subject, char *text, char *mimetype, char
            *filename, char *filename_alias, ios_send_email_cb cb, void *userdata)
    void ios_open_url(char *url)
    void load_url_webview(char *url, int x, int y, int width, int height)
    float ios_uiscreen_get_scale()
    int ios_uiscreen_get_dpi()
    padding ios_get_safe_area()

cdef void _send_email_done(char *status, void *data):
    cdef object callback = <object>data
    callback(status)
    Py_DECREF(callback)

#
#Support for iOS webview
#
class IOSWebView(object):
    def open(self, url, x, y, width, height):
        open_url_wbv(url, x, y, width, height)


def open_url_wbv(url, x, y, width, height):
    '''
    OPEN URL in webview

    :Parameters:
        `url`: str
            URL string

        `height`: int
            Height of the window

        `width`: int
            Width of the window

    Example for opening up a web page in WKWebview::

        import ios
        url = "http://www.google.com"
        ios.IOSWebView().open(url, x, y, width, height)
    '''
    load_url_webview(url, x, y, width, height)

#
# Support for webbrowser module
#

class IosBrowser(object):
    def open(self, url, new=0, autoraise=True):
        open_url(url)
    def open_new(self, url):
        open_url(url)
    def open_new_tab(self, url):
        open_url(url)

import webbrowser
try:
    # python 2
    webbrowser.register('ios', IosBrowser, None, -1)
except:
    # python 3
    webbrowser.register('ios', IosBrowser, None, preferred=True)
#
# API
#

__version__ = (1, 1, 0)

def open_url(url):
    '''Open an URL in Safari

    :Parameters:
        `url`: str
            The url string
    '''
    cdef char *j_url = NULL

    if url is not None:
        if type(url) is unicode:
            url = url.encode('UTF-8')
        j_url = <bytes>url

    ios_open_url(j_url)


def send_email(subject, text, mimetype=None, filename=None, filename_alias=None, callback=None):
    '''Send an email using the IOS api.

    :Parameters:
        `subject`: str
            Subject of the email
        `text`: str
            Content of the email
        `mimetype`: str
            Mimetype of the attachment if exist
        `filename`: str
            Full path of the filename to attach, must be used with mimetype.
        `filename_alias`: str
            Name of the file that will be shown to the user. If none is set, it
            will use the basename of filename.
        `callback`: func(status)
            Callback that can be called when the email interface have been
            removed. A status will be passed as the first argument: "cancelled",
            "saved", "sent", "failed", "unknown", "cannotsend".

    .. note::

        The application must have the window created to be able to use that
        method. Trying to send an email without the application running will
        crash.

    Example for sending a simple hello world::

        ios.send_email('This is my subject',
            'Hello you!\n\nThis is an hello world.')

    Send a mail with an attachment::

        from os.path import realpath
        ios.send_email('Mail with attachment',
            'Your attachment will be just after this message.',
            mimetype='image/png',
            filename=realpath('mylogo.png'))

    Getting the status of the mail with the callback

        from kivy.app import App

        class EmailApp(App):
            def callback_email(self, status):
                print 'The email have been', status

            def send_email(self, *largs):
                print 'Sending an email'
                ios.send_email('Hello subject', 'World body',
                    callback=self.callback_email)

            def build(self):
                btn = Button(text='Click me')
                btn.bind(on_release=self.send_email)
                return btn

        if __name__ == '__main__':
            EmailApp().run()

    '''
    cdef char *j_mimetype = NULL
    cdef char *j_filename = NULL
    cdef char *j_subject = NULL
    cdef char *j_text = NULL
    cdef char *j_title = NULL
    cdef char *j_filename_alias = NULL

    if subject is not None:
        if type(subject) is unicode:
            subject = subject.encode('UTF-8')
        j_subject = <bytes>subject
    if text is not None:
        if type(text) is unicode:
            text = text.encode('UTF-8')
        j_text = <bytes>text
    if mimetype is not None:
        j_mimetype = <bytes>mimetype
    if filename is not None:
        j_filename = <bytes>filename

        if filename_alias is None:
            filename_alias = basename(filename)
        if type(filename_alias) is unicode:
            filename_alias = filename_alias.encode('UTF-8')
        j_filename_alias = <bytes>filename_alias


    Py_INCREF(callback)

    ret = ios_send_email(j_subject, j_text, j_mimetype, j_filename,
            j_filename_alias, _send_email_done, <void *>callback)
    if ret == 0:
        callback('failed')
        return 0
    elif ret == -1:
        callback('cannotsend')
        return 0

    return 1

def get_scale():
    '''Return the UIScreen scale (1 on iPad, 2 on iPad 3)
    '''
    return ios_uiscreen_get_scale()

def get_dpi():
    '''Return the approximate DPI of the screen
    '''
    return ios_uiscreen_get_dpi()

def get_safe_area():
    '''Return the safe area bounds
    '''
    return ios_get_safe_area()


from pyobjus import autoclass, selector, protocol
from pyobjus.protocols import protocols

NSNotificationCenter = autoclass('NSNotificationCenter')

protocols["KeyboardDelegates"] = {
    'keyboardWillShow': ('v16@0:4@8', "v32@0:8@16"),
    'keyboardDidHide': ('v16@0:4@8', "v32@0:8@16")}


class IOSKeyboard(object):
    '''Get listener for keyboard height.
    '''

    kheight = 0

    def __init__(self, **kwargs):
        super(IOSKeyboard, self).__init__()
        NSNotificationCenter.defaultCenter().addObserver_selector_name_object_(self, selector("keyboardWillShow"), "UIKeyboardWillShowNotification", None)
        NSNotificationCenter.defaultCenter().addObserver_selector_name_object_(self, selector("keyboardDidHide"), "UIKeyboardDidHideNotification", None)

    @protocol('KeyboardDelegates')
    def keyboardWillShow(self, notification):
        self.kheight = get_scale() * notification.userInfo().objectForKey_(
            'UIKeyboardFrameEndUserInfoKey').CGRectValue().size.height
        from kivy.core.window import Window
        Window.trigger_keyboard_height()

    @protocol('KeyboardDelegates')
    def keyboardDidHide(self, notification):
        self.kheight = 0
        from kivy.core.window import Window
        Window.trigger_keyboard_height()

iOSKeyboard = IOSKeyboard()

def get_kheight():
    return iOSKeyboard.kheight



from kivy_ios.toolchain import Recipe, shprint
from os.path import join
import sh


class FreetypeRecipe(Recipe):
    version = "2.5.5"
    url = "http://download.savannah.gnu.org/releases/freetype/freetype-{version}.tar.bz2"
    library = "objs/.libs/libfreetype.a"
    include_dir = ["include", ("builds/unix/ftconfig.h", "config/ftconfig.h")]
    include_per_arch = True

    def build_arch(self, arch):
        build_env = arch.get_env()
        configure = sh.Command(join(self.build_dir, "configure"))
        shprint(configure,
                "CC={}".format(build_env["CC"]),
                "LD={}".format(build_env["LD"]),
                "CFLAGS={}".format(build_env["CFLAGS"]),
                "LDFLAGS={}".format(build_env["LDFLAGS"]),
                "--prefix=/",
                "--host={}".format(arch.triple),
                "--without-png",
                "--without-bzip2",
                "--without-fsspec",
                "--without-harfbuzz",
                "--without-old-mac-fonts",
                "--enable-static=yes",
                "--enable-shared=no")
        shprint(sh.make, "clean")
        shprint(sh.make, self.ctx.concurrent_make)


recipe = FreetypeRecipe()



from kivy_ios.toolchain import CythonRecipe
from os.path import join


class FFPyplayerRecipe(CythonRecipe):
    version = "4.2.0"
    url = "https://github.com/matham/ffpyplayer/archive/{version}.zip"
    library = "libffpyplayer.a"
    depends = ["python", "ffmpeg"]
    pbx_frameworks = [
        "CoreVideo", "CoreMedia", "CoreImage", "AVFoundation", "UIKit",
        "CoreMotion"]
    pbx_libraries = ["libiconv"]
    pre_build_ext = True

    def get_recipe_env(self, arch):
        env = super(FFPyplayerRecipe, self).get_recipe_env(arch)
        env["CC"] += " -I{}".format(
            join(self.ctx.dist_dir, "include", arch.arch, "libffi"))
        env["SDL_INCLUDE_DIR"] = join(self.ctx.dist_dir, "include",
                                      "common", "sdl2")
        env["FFMPEG_INCLUDE_DIR"] = join(self.ctx.dist_dir, "include",
                                         arch.arch, "ffmpeg")
        env["CONFIG_POSTPROC"] = "0"
        return env

    def prebuild_arch(self, arch):
        # common to all archs
        if self.has_marker("patched"):
            return
        self.apply_patch("misc-visibility.patch")
        self.set_marker("patched")


recipe = FFPyplayerRecipe()



from kivy_ios.toolchain import Recipe, shprint
from os.path import join
import sh
import logging

logger = logging.getLogger(__name__)


class HostOpensslRecipe(Recipe):
    version = "1.1.1g"
    url = "http://www.openssl.org/source/openssl-{version}.tar.gz"
    archs = ["x86_64"]

    def get_build_env(self):
        build_env = self.ctx.env.copy()
        self.build_env_x86_84 = build_env
        return build_env

    def build_x86_64(self):
        build_env = self.get_build_env()
        configure = sh.Command(join(self.build_dir, "Configure"))
        shprint(configure,
                "darwin64-x86_64-cc",
                _env=build_env)
        shprint(sh.make, "clean")
        shprint(sh.make, self.ctx.concurrent_make, "build_libs")

    def install(self):
        sh.mkdir('-p', join(self.ctx.dist_dir, 'hostopenssl'))
        sh.cp('-r', join(self.get_build_dir('x86_64'), 'include'),
              join(self.ctx.dist_dir, 'hostopenssl', 'include'))
        sh.mkdir('-p', join(self.ctx.dist_dir, 'hostopenssl', 'lib'))
        sh.cp(join(self.get_build_dir('x86_64'), 'libssl.a'),
              join(self.ctx.dist_dir, 'hostopenssl', 'lib'))
        sh.cp(join(self.get_build_dir('x86_64'), 'libcrypto.a'),
              join(self.ctx.dist_dir, 'hostopenssl', 'lib'))


recipe = HostOpensslRecipe()



"""
This module houses context managers to assist in the managing of state during
kivy-ios builds.
"""
from logging import getLogger
from contextlib import contextmanager
from os import getcwd, chdir, environ
from os.path import expanduser


logger = getLogger(__name__)


@contextmanager
def cd(newdir):
    """
    Set the current working directory to `newdir` for the duration of the
    context.
    """
    prevdir = getcwd()
    logger.info("cd {}".format(newdir))
    chdir(expanduser(newdir))
    try:
        yield
    finally:
        logger.info("cd {}".format(prevdir))
        chdir(prevdir)


@contextmanager
def python_path(newdir):
    """
    Set the PYTHONPATH environmnet variable to `newdir` for the duraiton of the
    context.
    """
    prevdir = environ.get("PYTHONPATH")
    logger.debug("Setting PYTHONPATH to {}".format(newdir))
    environ["PYTHONPATH"] = newdir
    try:
        yield
    finally:
        logger.debug("Setting PYTHONPATH to {}".format(prevdir))
        if prevdir is None:
            environ.pop("PYTHONPATH")
        else:
            environ["PYTHONPATH"] = prevdir



'''Recipe for pycrypto on ios
'''
from kivy_ios.toolchain import CythonRecipe, shprint
from kivy_ios.context_managers import cd
from os.path import join
import sh


class PycryptoRecipe(CythonRecipe):
    version = "2.6.1"
    url = "https://ftp.dlitz.net/pub/dlitz/crypto/pycrypto/pycrypto-{version}.tar.gz"
    depends = ["python", "openssl"]
    include_per_arch = True
    library = "libpycrypto.a"

    def build_arch(self, arch):
        build_env = arch.get_env()
        self.apply_patch('hash_SHA2_template.c.patch', target_dir=self.build_dir + '/src')
        configure = sh.Command(join(self.build_dir, "configure"))
        shprint(configure,
                "CC={}".format(build_env["CC"]),
                "LD={}".format(build_env["LD"]),
                "CFLAGS={}".format(build_env["CFLAGS"]),
                "LDFLAGS={} -Wno-error ".format(build_env["LDFLAGS"]),
                "--prefix=/",
                "--host={}".format(arch),
                "ac_cv_func_malloc_0_nonnull=yes",
                "ac_cv_func_realloc_0_nonnull=yes")
        super(PycryptoRecipe, self).build_arch(arch)

    def install(self):
        arch = list(self.filtered_archs)[0]
        build_dir = self.get_build_dir(arch.arch)
        hostpython = sh.Command(self.ctx.hostpython)
        build_env = arch.get_env()
        dest_dir = join(self.ctx.dist_dir, "root", "python")
        build_env['PYTHONPATH'] = join(dest_dir, 'lib', 'python3.7', 'site-packages')
        with cd(build_dir):
            shprint(hostpython, "setup.py", "install", "--prefix", dest_dir, _env=build_env)


recipe = PycryptoRecipe()



from kivy_ios.toolchain import Recipe, shprint
import sh
from os.path import exists


class LibffiRecipe(Recipe):
    version = "3.2.1"
    url = "https://sourceware.org/pub/libffi/libffi-{version}.tar.gz"
    library = "build/Release-{arch.sdk}/libffi.a"
    include_per_arch = True
    include_dir = "build_{arch.sdk}-{arch.arch}/include"
    include_name = "ffi"
    archs = ["x86_64", "arm64"]

    def prebuild_arch(self, arch):
        if self.has_marker("patched"):
            return
        # XCode 10 minimum is 8.0 now.
        shprint(sh.sed,
                "-i.bak",
                "s/-miphoneos-version-min=5.1.1/-miphoneos-version-min=8.0/g",
                "generate-darwin-source-and-headers.py")
        self.apply_patch("fix-win32-unreferenced-symbol.patch")
        self.apply_patch("generate-darwin-source-and-headers-python3-items.patch")
        self.set_marker("patched")

    def build_arch(self, arch):
        if exists("generate-darwin-source-and-headers.py"):
            shprint(
                sh.mv,
                "generate-darwin-source-and-headers.py",
                "_generate-darwin-source-and-headers.py")
            shprint(sh.touch, "generate-darwin-source-and-headers.py")
        python3 = sh.Command("python3")
        shprint(python3, "_generate-darwin-source-and-headers.py", "--only-ios")
        shprint(sh.xcodebuild, self.ctx.concurrent_xcodebuild,
                "ONLY_ACTIVE_ARCH=NO",
                "ARCHS={}".format(arch.arch),
                "-sdk", arch.sdk,
                "-project", "libffi.xcodeproj",
                "-target", "libffi-iOS",
                "-configuration", "Release")


recipe = LibffiRecipe()



from kivy_ios.toolchain import Recipe, shprint
from os.path import join
import sh


arch_mapper = {'i386': 'darwin-i386-cc',
               'x86_64': 'darwin64-x86_64-cc',
               'armv7': 'ios-cross',
               'arm64': 'ios64-cross'}


class OpensslRecipe(Recipe):
    version = "1.1.1g"
    url = "http://www.openssl.org/source/openssl-{version}.tar.gz"
    libraries = ["libssl.a", "libcrypto.a"]
    include_dir = "include"
    include_per_arch = True

    def build_arch(self, arch):
        build_env = arch.get_env()
        target = arch_mapper[arch.arch]
        shprint(sh.env, _env=build_env)
        sh.perl(join(self.build_dir, "Configure"),
                target,
                _env=build_env)
        if target.endswith('-cross'):
            with open('Makefile', 'r') as makefile:
                filedata = makefile.read()
            filedata = filedata.replace('$(CROSS_TOP)/SDKs/$(CROSS_SDK)', arch.sysroot)
            with open('Makefile', 'w') as makefile:
                makefile.write(filedata)
        shprint(sh.make, "clean")
        shprint(sh.make, self.ctx.concurrent_make, "build_libs")


recipe = OpensslRecipe()



from kivy_ios.toolchain import CythonRecipe


class IosRecipe(CythonRecipe):
    version = "master"
    url = "src"
    library = "libios.a"
    depends = ["python"]
    pbx_frameworks = ["MessageUI", "CoreMotion", "UIKit", "WebKit", "Photos"]

    def install(self):
        self.install_python_package(
            name=self.so_filename("ios"), is_dir=False)


recipe = IosRecipe()



import os
from kivy_ios.toolchain import Recipe
from os.path import join
import sh
import fnmatch
from distutils.dir_util import copy_tree


class ZbarLightRecipe(Recipe):
    version = '1.2'
    url = 'https://github.com/Polyconseil/zbarlight/archive/{version}.tar.gz'
    library = "zbarlight.a"
    depends = ['hostpython3', 'python3', 'libzbar']
    pbx_libraries = ["libz", "libbz2", 'libc++', 'libsqlite3', 'CoreMotion']
    include_per_arch = True

    def get_zbar_env(self, arch):
        build_env = arch.get_env()
        dest_dir = join(self.ctx.dist_dir, "root", "python")
        build_env["IOSROOT"] = self.ctx.root_dir
        build_env["IOSSDKROOT"] = arch.sysroot
        build_env["LDSHARED"] = join(self.ctx.root_dir, "tools", "liblink")
        build_env["ARM_LD"] = build_env["LD"]
        build_env["ARCH"] = arch.arch
        build_env["C_INCLUDE_PATH"] = join(arch.sysroot, "usr", "include")
        build_env["LIBRARY_PATH"] = join(arch.sysroot, "usr", "lib")
        build_env['PYTHONPATH'] = join(dest_dir, 'lib', 'python3.7', 'site-packages')
        build_env["CFLAGS"] = " ".join([
            " -I{}".format(join(self.ctx.dist_dir, "include", arch.arch, "libzbar", 'zbar')) +
            " -arch {}".format(arch.arch)
            ])
        build_env['LDFLAGS'] += " -lios -lpython -lzbar"
        return build_env

    def build_arch(self, arch):
        build_env = self.get_zbar_env(arch)
        hostpython = sh.Command(self.ctx.hostpython)
        shprint(hostpython, "setup.py", "build",   # noqa: F821
                _env=build_env)
        self.apply_patch("zbarlight_1_2.patch")  # Issue getting the version, hard coding for now
        self.biglink()

    def install(self):
        arch = list(self.filtered_archs)[0]
        build_dir = join(self.get_build_dir(arch.arch), 'build',
                         'lib.macosx-10.13-x86_64-2.7', 'zbarlight')
        dist_dir = join(self.ctx.dist_dir, 'root', 'python3', 'lib',
                        'python3.7', 'site-packages', 'zbarlight')
        # Patch before Copying
        # self.apply_patch("zbarlight_1_2.patch")#Issue getting the version, hard coding for now
        copy_tree(build_dir, dist_dir)
        os.remove(join(dist_dir, '_zbarlight.c'))

    def _patch__init__(self):
        init = join(self.ctx.dist_dir, 'root', 'python3', 'lib', 'python3.7',
                    'site-packages', 'zbarlight', "__init__.py")
        shprint(  # noqa: F821
            sh.sed, "-i.bak",
            "s/__version__ = pkg_resources.get_distribution('zbarlight').version'"
            "/__version__ = '{version}'/g",
            init)

    def biglink(self):
        dirs = []
        for root, dirnames, filenames in os.walk(self.build_dir):
            if fnmatch.filter(filenames, "*.so.libs"):
                dirs.append(root)
        cmd = sh.Command(join(self.ctx.root_dir, "tools", "biglink"))
        shprint(cmd, join(self.build_dir, "zbarlight.a"), *dirs)  # noqa: F821


recipe = ZbarLightRecipe()



# -*- coding: utf-8 -*-
from kivy_ios.toolchain import Recipe, shprint
from os.path import join
import sh


class PngRecipe(Recipe):
    version = '1.6.26'
    url = 'http://downloads.sourceforge.net/sourceforge/libpng/libpng-{version}.tar.gz'
    depends = ["python"]
    library = '.libs/libpng16.a'

    def build_arch(self, arch):
        build_env = arch.get_env()
        configure = sh.Command(join(self.build_dir, "configure"))
        shprint(configure,
                "CC={}".format(build_env["CC"]),
                "LD={}".format(build_env["LD"]),
                "CFLAGS={}".format(build_env["CFLAGS"]),
                "LDFLAGS={}".format(build_env["LDFLAGS"]),
                "--prefix=/",
                "--host={}".format(arch.triple),
                "--disable-shared")
        shprint(sh.make, "clean")
        shprint(sh.make, self.ctx.concurrent_make, _env=build_env)


recipe = PngRecipe()



"""
Icon and LaunchImage generator for iOS
======================================

.. author:: Mathieu Virbel <mat@meltingrocks.com>
"""
# flake8: noqa (E121 mainly)

__all__ = ["launchimage"]

import sh
import json
from PIL import Image
from os.path import join, exists
from os import makedirs

appicon_json = {
  "images": [
    {
      "idiom": "iphone",
      "size": "20x20",
      "scale": "2x",
      "filename": "Icon40.png"
    },
    {
      "idiom": "iphone",
      "size": "20x20",
      "scale": "3x",
      "filename": "Icon60.png"
    },
    {
      "size": "29x29",
      "idiom": "iphone",
      "filename": "Icon29.png",
      "scale": "1x"
    },
    {
      "size": "29x29",
      "idiom": "iphone",
      "filename": "Icon58.png",
      "scale": "2x"
    },
    {
      "size": "29x29",
      "idiom": "iphone",
      "filename": "Icon87.png",
      "scale": "3x"
    },
    {
      "size": "40x40",
      "idiom": "iphone",
      "filename": "Icon80.png",
      "scale": "2x"
    },
    {
      "size": "40x40",
      "idiom": "iphone",
      "filename": "Icon120.png",
      "scale": "3x"
    },
    {
      "size": "57x57",
      "idiom": "iphone",
      "filename": "Icon57.png",
      "scale": "1x"
    },
    {
      "size": "57x57",
      "idiom": "iphone",
      "filename": "Icon114.png",
      "scale": "2x"
    },
    {
      "size": "60x60",
      "idiom": "iphone",
      "filename": "Icon120.png",
      "scale": "2x"
    },
    {
      "size": "60x60",
      "idiom": "iphone",
      "filename": "Icon180.png",
      "scale": "3x"
    },
    {
      "idiom": "ipad",
      "size": "20x20",
      "filename": "Icon20.png",
      "scale": "1x"
    },
    {
      "idiom": "ipad",
      "size": "20x20",
      "filename": "Icon40.png",
      "scale": "2x"
    },
    {
      "size": "29x29",
      "idiom": "ipad",
      "filename": "Icon29.png",
      "scale": "1x"
    },
    {
      "size": "29x29",
      "idiom": "ipad",
      "filename": "Icon58.png",
      "scale": "2x"
    },
    {
      "size": "40x40",
      "idiom": "ipad",
      "filename": "Icon40.png",
      "scale": "1x"
    },
    {
      "size": "40x40",
      "idiom": "ipad",
      "filename": "Icon80.png",
      "scale": "2x"
    },
    {
      "size": "50x50",
      "idiom": "ipad",
      "filename": "Icon50.png",
      "scale": "1x"
    },
    {
      "size": "50x50",
      "idiom": "ipad",
      "filename": "Icon100.png",
      "scale": "2x"
    },
    {
      "size": "72x72",
      "idiom": "ipad",
      "filename": "Icon72.png",
      "scale": "1x"
    },
    {
      "size": "72x72",
      "idiom": "ipad",
      "filename": "Icon144.png",
      "scale": "2x"
    },
    {
      "size": "76x76",
      "idiom": "ipad",
      "filename": "Icon76.png",
      "scale": "1x"
    },
    {
      "size": "76x76",
      "idiom": "ipad",
      "filename": "Icon152.png",
      "scale": "2x"
    },
    # If activated, we got a submission error:
    # "Error ITMS-9000: Invalid Image Path - No image found at the path
    # referenced under key 'CFBundleIcons': 'AppIcon120x120'"
    # {
    #   "size": "120x120",
    #   "idiom": "car",
    #   "filename": "Icon120.png",
    #   "scale": "1x"
    # },
    {
      "size": "24x24",
      "idiom": "watch",
      "scale": "2x",
      "filename": "Icon48.png",
      "role": "notificationCenter",
      "subtype": "38mm"
    },
    {
      "size": "27.5x27.5",
      "idiom": "watch",
      "scale": "2x",
      "filename": "Icon55.png",
      "role": "notificationCenter",
      "subtype": "42mm"
    },
    {
      "size": "29x29",
      "idiom": "watch",
      "filename": "Icon58.png",
      "role": "companionSettings",
      "scale": "2x"
    },
    {
      "size": "29x29",
      "idiom": "watch",
      "filename": "Icon87.png",
      "role": "companionSettings",
      "scale": "3x"
    },
    {
      "size": "40x40",
      "idiom": "watch",
      "scale": "2x",
      "filename": "Icon80.png",
      "role": "appLauncher",
      "subtype": "38mm"
    },
    {
      "size": "44x44",
      "idiom": "watch",
      "scale": "2x",
      "filename": "Icon88.png",
      "role": "longLook",
      "subtype": "42mm"
    },
    {
      "size": "86x86",
      "idiom": "watch",
      "scale": "2x",
      "filename": "Icon172.png",
      "role": "quickLook",
      "subtype": "38mm"
    },
    {
      "size": "44x44",
      "idiom": "watch",
      "scale": "2x",
      "filename": "Icon88.png",
      "role": "appLauncher",
      "subtype": "40mm"
    },
    {
      "size": "50x50",
      "idiom": "watch",
      "scale": "2x",
      "filename": "Icon100.png",
      "role": "appLauncher",
      "subtype": "44mm"
    },
    {
      "size": "98x98",
      "idiom": "watch",
      "scale": "2x",
      "filename": "Icon196.png",
      "role": "quickLook",
      "subtype": "42mm"
    },
    {
      "size": "108x108",
      "idiom": "watch",
      "scale": "2x",
      "filename": "Icon216.png",
      "role": "quickLook",
      "subtype": "44mm"
    },
    {
      "size": "16x16",
      "idiom": "mac",
      "filename": "Icon16.png",
      "scale": "1x"
    },
    {
      "size": "16x16",
      "idiom": "mac",
      "filename": "Icon32.png",
      "scale": "2x"
    },
    {
      "size": "32x32",
      "idiom": "mac",
      "filename": "Icon32.png",
      "scale": "1x"
    },
    {
      "size": "32x32",
      "idiom": "mac",
      "filename": "Icon64.png",
      "scale": "2x"
    },
    {
      "size": "128x128",
      "idiom": "mac",
      "filename": "Icon128.png",
      "scale": "1x"
    },
    {
      "size": "128x128",
      "idiom": "mac",
      "filename": "Icon256.png",
      "scale": "2x"
    },
    {
      "size": "256x256",
      "idiom": "mac",
      "filename": "Icon256.png",
      "scale": "1x"
    },
    {
      "size": "256x256",
      "idiom": "mac",
      "filename": "Icon512.png",
      "scale": "2x"
    },
    {
      "size": "512x512",
      "idiom": "mac",
      "filename": "Icon512.png",
      "scale": "1x"
    },
    {
      "size": "512x512",
      "idiom": "mac",
      "filename": "Icon1024.png",
      "scale": "2x"
    },
    {
      "size": "83.5x83.5",
      "idiom": "ipad",
      "filename": "Icon167.png",
      "scale": "2x"
    },
    {
      "idiom": "ios-marketing",
      "size": "1024x1024",
      "scale": "1x",
      "filename": "Icon1024.png"
    },
    {
      "idiom": "watch-marketing",
      "size": "1024x1024",
      "scale": "1x",
      "filename": "Icon1024.png"
    },
  ],
  "info": {
    "version": 1,
    "author": "xcode"
  },
  # "properties": {
  #   "pre-rendered": True
  # }
}


launchimage_json = {
  "images": [
    {
      "extent": "full-screen",
      "idiom": "iphone",
      "subtype": "736h",
      "filename": "Default1242x2208.png",
      "minimum-system-version": "8.0",
      "orientation": "portrait",
      "scale": "3x"
    },
    {
      "extent": "full-screen",
      "idiom": "iphone",
      "subtype": "736h",
      "filename": "Default2208x1242.png",
      "minimum-system-version": "8.0",
      "orientation": "landscape",
      "scale": "3x"
    },
    {
      "extent": "full-screen",
      "idiom": "iphone",
      "subtype": "667h",
      "filename": "Default750x1334.png",
      "minimum-system-version": "8.0",
      "orientation": "portrait",
      "scale": "2x"
    },
    {
      "orientation": "portrait",
      "idiom": "iphone",
      "extent": "full-screen",
      "minimum-system-version": "7.0",
      "filename": "Default640x960.png",
      "scale": "2x"
    },
    {
      "extent": "full-screen",
      "idiom": "iphone",
      "subtype": "retina4",
      "filename": "Default640x1136.png",
      "minimum-system-version": "7.0",
      "orientation": "portrait",
      "scale": "2x"
    },
    {
      "orientation": "portrait",
      "idiom": "ipad",
      "extent": "full-screen",
      "minimum-system-version": "7.0",
      "filename": "Default768x1024.png",
      "scale": "1x"
    },
    {
      "orientation": "landscape",
      "idiom": "ipad",
      "extent": "full-screen",
      "minimum-system-version": "7.0",
      "filename": "Default1024x768.png",
      "scale": "1x"
    },
    {
      "orientation": "portrait",
      "idiom": "ipad",
      "extent": "full-screen",
      "minimum-system-version": "7.0",
      "filename": "Default1536x2048.png",
      "scale": "2x"
    },
    {
      "orientation": "landscape",
      "idiom": "ipad",
      "extent": "full-screen",
      "minimum-system-version": "7.0",
      "filename": "Default2048x1536.png",
      "scale": "2x"
    },
    {
      "orientation": "portrait",
      "idiom": "iphone",
      "extent": "full-screen",
      "filename": "Default320x480.png",
      "scale": "1x"
    },
    {
      "orientation": "portrait",
      "idiom": "iphone",
      "extent": "full-screen",
      "filename": "Default640x960.png",
      "scale": "2x"
    },
    {
      "orientation": "portrait",
      "idiom": "iphone",
      "extent": "full-screen",
      "filename": "Default640x1136.png",
      "subtype": "retina4",
      "scale": "2x"
    },
    {
      "orientation": "portrait",
      "idiom": "ipad",
      "extent": "full-screen",
      "filename": "Default768x1024.png",
      "scale": "1x"
    },
    {
      "orientation": "landscape",
      "idiom": "ipad",
      "extent": "full-screen",
      "filename": "Default1024x768.png",
      "scale": "1x"
    },
    {
      "orientation": "portrait",
      "idiom": "ipad",
      "extent": "full-screen",
      "filename": "Default1536x2048.png",
      "scale": "2x"
    },
    {
      "orientation": "landscape",
      "idiom": "ipad",
      "extent": "full-screen",
      "filename": "Default2048x1536.png",
      "scale": "2x"
    },
  ],
  "info": {
    "version": 1,
    "author": "xcode"
  }
}


def icon(image_xcassets, image_fn):
    """Generate all the possible Icon from a single image_fn
    """
    appicon_dir = join(image_xcassets, "AppIcon.appiconset")
    if not exists(appicon_dir):
        makedirs(appicon_dir)
    with open(join(appicon_dir, "Contents.json"), "w") as fd:
        json.dump(appicon_json, fd)

    options = (
        # iPhone
        # Spotlight - iOS 5,6
        # Settings - iOS 5-8
        # 29pt - 1x,2x,3x
        ("87", None, "Icon87.png"),
        ("58", None, "Icon58.png"),
        ("29", "Icon58.png", "Icon29.png"),

        # iPhone notification
        # 20pt - 2x,3x
        # ("40", None, "Icon40.png"),
        ("60", None, "Icon60.png"),

        # iPhone
        # Spotlight - iOS 7-8
        # 40pt 2x,3x
        ("120", None, "Icon120.png"),
        ("80", None, "Icon80.png"),

        # iPhone
        # App - iOS 5,6
        # 57pt 1x,2x
        ("114", None, "Icon114.png"),
        ("57", "Icon114.png", "Icon57.png"),

        # iPhone
        # App - iOS 7,8
        # 60pt 2x,3x
        ("180", None, "Icon180.png"),
        # ("120", None, "Icon120.png # duplicate"),

        # iPad
        # Notifications
        # 20pt 1x,2x
        ("20", "Icon80.png", "Icon20.png"),
        ("40", "Icon80.png", "Icon40.png"),

        # iPad
        # Settings iOS 5-8
        # ("58", None, "Icon58.png # duplicate"),
        # ("29", "Icon58.png", "Icon29.png # duplicate"),

        # iPad
        # Spotlight iOS 7,8
        # 40pt 1x,2x
        # ("80", None, "Icon80.png # duplicate"),
        ("40", "Icon80.png", "Icon40.png"),

        # iPad
        # Spotlight iOS 5,6
        # 50pt 1x,2x
        ("100", None, "Icon100.png"),
        ("50", "Icon100.png", "Icon50.png"),

        # iPad
        # App iOS 5,6
        # 72pt 1x,2x
        ("144", None, "Icon144.png"),
        ("72", "Icon144.png", "Icon72.png"),

        # iPad
        # App iOS 7,8
        # 76pt 1x,2x
        ("152", None, "Icon152.png"),
        ("76", "Icon152.png", "Icon76.png"),

        # iPad
        # App iOS 9
        # 83.5pt 2x
        ("167", None, "Icon167.png"),


        # CarPlay
        # App iOS 8
        # 120pt 1x
        # ("120", None, "Icon120.png # duplicate"),


        # Apple Watch
        # Notification Center
        # 38mm, 42mm
        ("48", None, "Icon48.png"),
        ("55", None, "Icon55.png"),

        # Apple Watch
        # Companion Settings
        # 29pt 2x,3x
        # ("58", None, "Icon58.png # duplicate"),
        # ("87", None, "Icon87.png # duplicate"),

        # Apple Watch
        # Home Screen (All)
        # Long Look (38mm)
        # ("80", None, "Icon80.png # duplicate"),

        # Apple Watch
        # Long Look (42mm)
        ("88", None, "Icon88.png"),

        # Apple Watch
        # Short Look
        # 38mm, 42mm, 44mm
        ("172", None, "Icon172.png"),
        ("196", None, "Icon196.png"),
        ("216", None, "Icon216.png"),


        # OS X
        # 512pt 1x,2x
        ("1024", None, "Icon1024.png"),
        ("512", "Icon1024.png", "Icon512.png"),

        # OS X
        # 256pt 1x,2x
        # ("512", "Icon1024.png", "Icon512.png # duplicate"),
        ("256", "Icon512.png", "Icon256.png"),

        # OS X
        # 128pt 1x,2x
        # ("256", "Icon512.png", "Icon256.png # duplicate"),
        ("128", "Icon256.png", "Icon128.png"),

        # OS X
        # 32pt 1x,2x
        ("64", "Icon128.png", "Icon64.png"),
        ("32", "Icon64.png", "Icon32.png"),

        # OS X
        # 16pt 1x,2x
        # ("32", "Icon64.png", "Icon32.png # duplicate"),
        ("16", "Icon32.png", "Icon16.png"))

    _generate("AppIcon.appiconset", image_xcassets, image_fn, options, icon=True)


def launchimage(image_xcassets, image_fn):
    """Generate all the possible Launch Images from a single image_fn
    """
    launchimage_dir = join(image_xcassets, "LaunchImage.launchimage")
    if not exists(launchimage_dir):
        makedirs(launchimage_dir)
    with open(join(launchimage_dir, "Contents.json"), "w") as fd:
        json.dump(launchimage_json, fd)

    options = (
        # size, input, output
        # iPhone 3.5" @2x
        ("640 960", None, "Default640x960.png"),
        # iPhone 3.5" @1x
        ("320 480", None, "Default320x480.png"),
        # iPhone 4.0" @2x
        ("640 1136", None, "Default640x1136.png"),
        # iPhone 5.5" @3x - landscape
        ("2208 1242", None, "Default2208x1242.png"),
        # iPhone 5.5" @3x - portrait
        ("1242 2208", None, "Default1242x2208.png"),
        # iPhone 4.7" @2x
        ("750 1334", None, "Default750x1334.png"),
        # iPad @2x - landscape
        ("2048 1536", None, "Default2048x1536.png"),
        # iPad @2x - portrait
        ("1536 2048", None, "Default1536x2048.png"),
        # iPad @1x - landscape
        ("1024 768", None, "Default1024x768.png"),
        # iPad @1x - portrait
        ("768 1024", None, "Default768x1024.png"),
    )

    _generate("LaunchImage.launchimage", image_xcassets, image_fn, options)


def _buildimage(in_fn, out_fn, size, padcolor=None):
    im = Image.open(in_fn)

    # read the first left/bottom pixel
    bgcolor = im.getpixel((0, 0))

    # ensure the image fit in the destination size
    if im.size[0] > size[0] or im.size[1] > size[1]:
        f = max(im.size[0] / size[0], im.size[1] / size[1])
        newsize = int(im.size[0] / f), int(im.size[1] / f)
        im = im.resize(newsize)

    # create final image
    outim = Image.new("RGB", size, bgcolor[:3])
    x = (size[0] - im.size[0]) // 2
    y = (size[1] - im.size[1]) // 2
    outim.paste(im, (x, y))

    # save the image
    outim.save(out_fn)


def _generate(d, image_xcassets, image_fn, options, icon=False):
    for c, in_fn, out_fn in options:
        args = []
        if in_fn is not None:
            filename = join(image_xcassets, d, in_fn)
        else:
            filename = image_fn

        if icon:
            args += ["-Z", c]
            args += [
                "--out",
                join(image_xcassets, d, out_fn)
            ]
            print("sips", " ".join(args))
            sh.sips(*args)
        else:
            size = [int(x) for x in c.split()]
            _buildimage(filename, join(image_xcassets, d, out_fn), size)



import os
from glob import glob
from setuptools import setup, find_packages


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as f:
        return f.read()


def recursive_include(module):
    module_path = module.replace(".", "/") + "/"
    files = glob(f"{module_path}**", recursive=True)
    return [file.replace(module_path, "") for file in files]


setup(
    name="kivy-ios",
    version="1.2.0",
    description="Kivy for iOS",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    author="The Kivy team",
    author_email="kivy-dev@googlegroups.com",
    url="https://github.com/kivy/kivy-ios",
    python_requires=">=3.6.0",
    install_requires=["cookiecutter", "pbxproj", "Pillow", "requests", "sh"],
    packages=find_packages(),
    package_data={
        # note this method is a bit excessive as it includes absolutely everything
        # make sure you run with from a clean directory
        "kivy_ios": recursive_include("kivy_ios"),
    },
    entry_points={"console_scripts": ["toolchain = kivy_ios.toolchain:main"]},
)



from kivy_ios.toolchain import Recipe, shprint
from os.path import join
import sh


class LibSDL2ImageRecipe(Recipe):
    version = "2.0.4"
    url = "https://www.libsdl.org/projects/SDL_image/release/SDL2_image-{version}.tar.gz"
    library = "Xcode-iOS/build/Release-{arch.sdk}/libSDL2_image.a"
    include_dir = "SDL_image.h"
    depends = ["sdl2"]
    pbx_frameworks = ["CoreGraphics", "MobileCoreServices"]

    def build_arch(self, arch):
        shprint(sh.xcodebuild, self.ctx.concurrent_xcodebuild,
                "ONLY_ACTIVE_ARCH=NO",
                "ARCHS={}".format(arch.arch),
                "HEADER_SEARCH_PATHS={}".format(
                    join(self.ctx.include_dir, "common", "sdl2")),
                "-sdk", arch.sdk,
                "-project", "Xcode-iOS/SDL_image.xcodeproj",
                "-target", "libSDL_image-iOS",
                "-configuration", "Release")


recipe = LibSDL2ImageRecipe()



from kivy_ios.toolchain import Recipe, shprint
from os.path import join
import sh
import os
import fnmatch


class PillowRecipe(Recipe):
    version = "6.1.0"
    url = "https://pypi.python.org/packages/source/P/Pillow/Pillow-{version}.tar.gz"
    library = "libpillow.a"
    depends = ["hostpython3", "host_setuptools3", "freetype", "libjpeg", "python3", "ios"]
    python_depends = ["setuptools"]
    pbx_libraries = ["libz", "libbz2"]
    include_per_arch = True

    def get_pil_env(self, arch):
        build_env = arch.get_env()
        build_env["IOSROOT"] = self.ctx.root_dir
        build_env["IOSSDKROOT"] = arch.sysroot
        build_env["LDSHARED"] = join(self.ctx.root_dir, "tools", "liblink")
        build_env["ARM_LD"] = build_env["LD"]
        build_env["ARCH"] = arch.arch
        build_env["C_INCLUDE_PATH"] = join(arch.sysroot, "usr", "include")
        build_env["LIBRARY_PATH"] = join(arch.sysroot, "usr", "lib")
        build_env["CFLAGS"] = " ".join([
            "-I{}".format(join(self.ctx.dist_dir, "include", arch.arch, "freetype")) +
            " -I{}".format(join(self.ctx.dist_dir, "include", arch.arch, "libjpeg")) +
            " -arch {}".format(arch.arch)
            ])
        build_env['PATH'] = os.environ['PATH']
        return build_env

    def build_arch(self, arch):
        build_env = self.get_pil_env(arch)
        hostpython3 = sh.Command(self.ctx.hostpython)
        shprint(hostpython3, "setup.py", "build_ext", "--disable-tiff",
                "--disable-webp", "-g", _env=build_env)
        self.biglink()

    def install(self):
        arch = list(self.filtered_archs)[0]
        build_dir = self.get_build_dir(arch.arch)
        os.chdir(build_dir)
        hostpython3 = sh.Command(self.ctx.hostpython)
        build_env = self.get_pil_env(arch)
        dest_dir = join(self.ctx.dist_dir, "root", "python3")
        build_env['PYTHONPATH'] = join(dest_dir, 'lib', 'python3.8', 'site-packages')
        shprint(hostpython3, "setup.py", "install", "--prefix", dest_dir,
                _env=build_env)

    def biglink(self):
        dirs = []
        for root, dirnames, filenames in os.walk(self.build_dir):
            if fnmatch.filter(filenames, "*.so.libs"):
                dirs.append(root)
        cmd = sh.Command(join(self.ctx.root_dir, "tools", "biglink"))
        shprint(cmd, join(self.build_dir, "libpillow.a"), *dirs)


recipe = PillowRecipe()



from kivy_ios.toolchain import Recipe, shprint
from os.path import join
import sh
import shutil
import shlex


class LibSDL2TTFRecipe(Recipe):
    version = "2.0.14"
    url = "https://www.libsdl.org/projects/SDL_ttf/release/SDL2_ttf-{version}.tar.gz"
    library = "libSDL2_ttf.a"
    include_dir = "SDL_ttf.h"
    depends = ["sdl2", "freetype"]

    def build_arch(self, arch):
        # XCode-iOS have shipped freetype that don't work with i386
        # ./configure require too much things to setup it correcly.
        # so build by hand.
        build_env = arch.get_env()
        cc = sh.Command(build_env["CC"])
        output = join(self.build_dir, "SDL_ttf.o")
        args = shlex.split(build_env["CFLAGS"])
        args += ["-c", "-o", output, "SDL_ttf.c"]
        shprint(cc, *args)
        shprint(sh.ar, "-q", join(self.build_dir, "libSDL2_ttf.a"), output)

    def install(self):
        for arch in self.filtered_archs:
            shutil.copy(
                join(self.get_build_dir(arch.arch), "SDL_ttf.h"),
                join(self.ctx.include_dir, "common", "SDL2"))


recipe = LibSDL2TTFRecipe()



from kivy_ios.toolchain import Recipe, shprint
from os.path import join
import sh


class LibZBarRecipe(Recipe):

    version = '0.10'

    url = 'https://github.com/ZBar/ZBar/archive/{version}.zip'

    depends = ['hostpython3']

    library = 'zbar/.libs/libzbar.a'

    include_per_arch = True
    include_dir = [
        ("include", "")
        ]

    def prebuild_arch(self, arch):
        if self.has_marker("patched"):
            return
        self.apply_patch("werror.patch")
        self.set_marker("patched")

    def build_arch(self, arch):
        super(LibZBarRecipe, self).build_arch(arch)
        build_env = arch.get_env()
        build_env["CFLAGS"] = " ".join([
            "-I{}".format(join(self.ctx.dist_dir, "build", "libiconv", arch.arch)) +
            " -arch {}".format(arch.arch), build_env['CFLAGS']
            ])
        shprint(sh.Command('autoreconf'), '-vif')
        shprint(
            sh.Command('./configure'),
            "CC={}".format(build_env["CC"]),
            "LD={}".format(build_env["LD"]),
            "CFLAGS={}".format(build_env["CFLAGS"]),
            "LDFLAGS={}".format(build_env["LDFLAGS"]),
            "--host={}".format(arch.triple),
            '--target={}'.format(arch.triple),
            # Python bindings are compiled in a separated recipe
            '--with-python=no',
            '--with-gtk=no',
            '--with-qt=no',
            '--with-x=no',
            '--with-jpeg=no',
            '--with-imagemagick=no',
            '--enable-pthread=no',
            '--enable-video=no',
            "--disable-shared",
            _env=build_env)
        shprint(sh.make, 'clean')
        shprint(sh.make, _env=build_env)


recipe = LibZBarRecipe()



from kivy_ios.toolchain import Recipe, shprint
import sh


class LibSDL2Recipe(Recipe):
    # version = "2.0.9"
    # url = "https://www.libsdl.org/release/SDL2-{version}.tar.gz"
    version = "7cc4fc886d9e"
    url = "https://hg.libsdl.org/SDL/archive/{version}.tar.gz"
    library = "Xcode-iOS/SDL/build/Release-{arch.sdk}/libSDL2.a"
    include_dir = "include"
    pbx_frameworks = [
        "OpenGLES", "AudioToolbox", "QuartzCore", "CoreGraphics",
        "CoreMotion", "GameController", "AVFoundation", "Metal",
        "UIKit"]

    def prebuild_arch(self, arch):
        if self.has_marker("patched"):
            return
        self.apply_patch("uikit-transparent.patch")
        self.set_marker("patched")

    def build_arch(self, arch):
        env = arch.get_env()
        shprint(sh.xcodebuild, self.ctx.concurrent_xcodebuild,
                "ONLY_ACTIVE_ARCH=NO",
                "ARCHS={}".format(arch.arch),
                "CC={}".format(env['CC']),
                "-sdk", arch.sdk,
                "-project", "Xcode-iOS/SDL/SDL.xcodeproj",
                "-target", "libSDL-iOS",
                "-configuration", "Release")


recipe = LibSDL2Recipe()



from kivy_ios.toolchain import CythonRecipe


class PhotoRecipe(CythonRecipe):
    version = "master"
    url = "https://github.com/akshayaurora/photolibrary/archive/master.zip"
    library = "libphotolibrary.a"
    depends = ["python"]
    pbx_frameworks = ["UIKit", "Photos", "MobileCoreServices"]

    def install(self):
        self.install_python_package(name="photolibrary.so", is_dir=False)


recipe = PhotoRecipe()
